"""
Created on 15.09.2024

@author: wf
"""

from lodstorage.sparql import SPARQL


class Wikidata:
    @classmethod
    def get_sparql(cls):
        endpoint_uri = "https://query.wikidata.org/sparql"
        sparql = SPARQL(endpoint_uri)
        return sparql

    @classmethod
    def unprefix(self, qid: str):
        item_prefix = "http://www.wikidata.org/entity/"
        if qid.startswith(item_prefix):
            qid = qid.replace(item_prefix, "")
        return qid



"""
Created on 2024-08-15

@author: wf
"""

import sys

from ngwidgets.cmd import WebserverCmd

from genwiki.webserver import GenWikiWebServer


class GenWikiCmd(WebserverCmd):
    """
    command line handling for genealogy wiki frontend
    """

    def __init__(self):
        """
        constructor
        """
        config = GenWikiWebServer.get_config()
        WebserverCmd.__init__(self, config, GenWikiWebServer, DEBUG)


def main(argv: list = None):
    """
    main call
    """
    cmd = GenWikiCmd()
    exit_code = cmd.cmd_main(argv)
    return exit_code


DEBUG = 0
if __name__ == "__main__":
    if DEBUG:
        sys.argv.append("-d")
    sys.exit(main())



"""
Created on 2024-10-24

@author: wf
"""

from genwiki.playground import Playground
from tests.gbasetest import GenealogyBasetest


class TestPlayground(GenealogyBasetest):
    """
    test playground generation
    """

    def setUp(self, debug=True, profile=True):
        GenealogyBasetest.setUp(self, debug=debug, profile=profile)
        self.playground = Playground()

    def test_thumbnail_generation(self):
        """
        Test that the thumbnail URL is correctly generated from the original image URL
        with different sizes using subTest, and output the result for manual verification.
        """
        original_url = "https://upload.wikimedia.org/wikipedia/commons/b/b8/Aristaeus_Bosio_Louvre_LL51.jpg"
        test_sizes = [300, 131]

        for size in test_sizes:
            with self.subTest(size=size):
                thumbnail_url = self.playground.get_thumbnail_url(
                    original_url, size=size
                )
                # Debug output for manual verification
                if self.debug:
                    print(f"Generated thumbnail URL for size {size}: {thumbnail_url}")
                # Assert that the thumbnail URL contains the size
                self.assertIn(f"/{size}px-", thumbnail_url)
                self.assertTrue(
                    thumbnail_url.startswith(
                        "https://upload.wikimedia.org/wikipedia/commons/thumb/"
                    )
                )



"""
Created on 2024-08-15

@author: wf
"""

import json
import os

from lodstorage.params import Params
from lodstorage.query import QueryManager
from lodstorage.sql import SQLDB
from lodstorage.uml import UML
from ngwidgets.basetest import Basetest

from genwiki.convert import ParquetAdressbokToSql


class TestParquet(Basetest):
    """
    Test parquet handling using pyarrow
    """

    def setUp(self, debug=False, profile=True):
        Basetest.setUp(self, debug=debug, profile=profile)
        # Define the directory containing the Parquet files
        current_folder = os.path.dirname(__file__)
        self.genwiki_examples_folder = os.path.join(
            current_folder, "..", "genwiki_examples"
        )
        # Ensure the directory exists
        self.assertTrue(
            os.path.exists(self.genwiki_examples_folder),
            f"Directory not found: {self.genwiki_examples_folder}",
        )

    def get_db(self) -> SQLDB:
        debug = self.debug
        db_path = "/tmp/address.db"
        if not os.path.exists(db_path):
            pats = ParquetAdressbokToSql(folder=self.genwiki_examples_folder)
            parquet_data = pats.read()

            # Ensure there are parquet files to read
            self.assertTrue(
                parquet_data,
                f"No .parquet files found in {self.genwiki_examples_folder}",
            )

            pats.inject_year()
            # Convert to SQLite with the correct table name and column mapping
            sql_db = SQLDB(db_path)
            pats.convert(sql_db)
        else:
            sql_db = SQLDB(db_path)
        return sql_db

    def test_parquet(self):
        """
        Test reading parquet files using pyarrow and converting to SQLite
        """
        debug = self.debug
        debug = True
        sql_db = self.get_db()

        # Check the address table
        table_info = sql_db.getTableDict()
        if debug:
            print(json.dumps(table_info, indent=2))
        for p_name in ["weimarTH1851", "weimarTH1853"]:
            self.assertTrue(p_name in table_info)
        table_list = sql_db.getTableList()
        uml = UML()
        markup = uml.tableListToPlantUml(
            table_list,
            "Adressbücher",
            "de.compgen.genwiki2024",
            "address",
            withSkin=True,
        )
        print(markup)

    def test_queries(self):
        """
        Test SQL queries loaded from a YAML file and executed against the SQLite database.
        """
        debug = self.debug
        debug = True
        sql_db = self.get_db()
        yaml_path = os.path.join(self.genwiki_examples_folder, "queries.yaml")
        # Initialize the QueryManager
        test_params = {
            "Gesamtanzahl": {},
            "Stichprobe": {"limit": "5"},
            "BerufStatistik": {"limit": "10"},
            "PersonenSuche": {"suchbegriff": "Ziegler", "limit": "10"},
            "StraßenStatistik": {"straße": "Frauent", "limit": "50"},
            "StraßenVergleich": {"straße": "Frauent"},
        }
        qm = QueryManager(
            lang="sql", queriesPath=yaml_path, with_default=False, debug=self.debug
        )
        for i, (qname, query) in enumerate(qm.queriesByName.items()):
            query.database = "sqlite"
            self.assertTrue(qname in test_params, qname)
            params_dict = test_params[qname]
            params = Params(query.query)
            query.query = params.apply_parameters_with_check(params_dict)
            if debug:
                print(f"{i:3}:{query}")
            lod = sql_db.query(query.query)
            if debug:
                print(json.dumps(lod, indent=2))



"""
Created on 25.08.2024

@author: wf
"""

import json
import os
from dataclasses import dataclass, field
from datetime import datetime
from typing import List, Optional

import requests


@dataclass
class Position:
    lon: float
    lat: float
    height: int
    type: str


@dataclass
class ExternalReference:
    value: str


@dataclass
class Name:
    lang: str
    value: str


@dataclass
class Type:
    value: str


@dataclass
class Timespan:
    begin: dict
    end: dict


@dataclass
class Source:
    ref: Optional[str] = None
    note: Optional[str] = None


@dataclass
class Population:
    value: str
    timespan: Optional[Timespan] = None
    source: Optional[List[Source]] = None
    year: Optional[int] = None


@dataclass
class PostalCode:
    value: str
    timespan: Optional[Timespan] = None


@dataclass
class MunicipalId:
    value: str


@dataclass
class PartOf:
    ref: str
    timespan: Optional[Timespan] = None
    source: Optional[List[Source]] = None


@dataclass
class GOVObject:
    id: str
    lastModification: datetime
    position: Position
    externalReference: List[ExternalReference]
    name: List[Name]
    type: List[Type]
    population: List[Population]
    postalCode: List[PostalCode]
    municipalId: List[MunicipalId]
    partOf: List[PartOf]


class GOV_API:
    """
    Access to Webservice https://wiki.genealogy.net/GOV/Webservice
    with local caching
    """

    def __init__(self):
        self.url = "https://gov.genealogy.net/api/getObject"
        self.cache_dir = os.path.expanduser("~/.govapi")
        os.makedirs(self.cache_dir, exist_ok=True)

    def get_cache_path(self, gov_id: str) -> str:
        return os.path.join(self.cache_dir, f"{gov_id}.json")

    def get_from_cache(self, gov_id: str) -> Optional[dict]:
        cache_path = self.get_cache_path(gov_id)
        if os.path.exists(cache_path):
            with open(cache_path, "r") as f:
                return json.load(f)
        return None

    def save_to_cache(self, gov_id: str, data: dict):
        cache_path = self.get_cache_path(gov_id)
        with open(cache_path, "w") as f:
            json.dump(data, f)

    def get_raw_gov_object(self, gov_id: str):
        # Check cache first
        cached_data = self.get_from_cache(gov_id)
        if cached_data:
            return cached_data

        # If not in cache, fetch from API
        params = {"itemId": gov_id}
        response = requests.get(self.url, params=params)
        response.raise_for_status()
        data = response.json()

        # Save to cache
        self.save_to_cache(gov_id, data)

        return data

    def get_gov_object(self, gov_id: str) -> GOVObject:
        data = self.get_raw_gov_object(gov_id)
        # Convert lastModification to datetime
        data["lastModification"] = datetime.fromisoformat(
            data["lastModification"].replace("Z", "+00:00")
        )

        # Create GOVObject instance
        gov_object = GOVObject(**data)
        return gov_object



"""
Created on 2025-02-25

@author: wf
"""

import logging
import os
import shutil
import sys
import tarfile
import tempfile
from concurrent.futures import Future, ThreadPoolExecutor
from dataclasses import dataclass, field
from pathlib import Path
from typing import Generator, List, Optional, Tuple

import cairo
import djvu.decode
import numpy
from ngwidgets.profiler import Profiler

from genwiki.djvu_core import DjVuImage


@dataclass
class ImageJob:
    """
    Represents a processed DjVu page, including document, page, page job, and image data.
    """
    djvu_path: str# fully qualifying path of the container DjVu document
    document: djvu.decode.Document
    page: djvu.decode.Page
    page_index: int  # Added page_index to track position
    relurl: str  # Added relurl for context
    pagejob: Optional[djvu.decode.PageJob] = field(default=None)
    image: Optional[DjVuImage] = field(default=None)
    verbose: bool = False
    debug: bool = False
    # keep track of any errors
    error: Optional[Exception] = field(default=None)

    def __post_init__(self):
        """Initialize profiler if not provided"""
        self.profiler = Profiler(
            f"Image Job {self.relurl}#{self.page_index:04d}",
            profile=self.verbose or self.debug,
        )
        self.profiler.start()

    def log(self, msg):
        if self.verbose or self.debug:
            self.profiler.time(" " + msg)

    def get_size(self) -> Tuple[int, int]:
        """Get the width and height of the page if pagejob is available"""
        if self.pagejob:
            return self.pagejob.size
        return (0, 0)

    @staticmethod
    def get_prefix(relurl: str):
        prefix = os.path.splitext(os.path.basename(relurl))[0]
        return prefix

    @property
    def prefix(self) -> str:
        prefix = ImageJob.get_prefix(relurl=self.relurl)
        return prefix

    @property
    def filename(self) -> str:
        try:
            # Attempt to safely decode the file name
            filename = self.page.file.name.encode(
                "utf-8", errors="replace"
            ).decode("utf-8")
        except Exception as e:
            if self.debug:
                logging.warn(
                    f"Failed to decode filename for page {self.page_index}: {e}"
                )
            filename = f"page_{self.page_index:04d}.djvu"
        return filename

    @property
    def dirname(self)->str:
        dirname=os.path.dirname(self.djvu_path)
        return dirname

    @property
    def filepath(self)->str:
        filepath=os.path.join(self.dirname,self.filename)
        return filepath

class DjVuContext(djvu.decode.Context):
    """
    A lightweight wrapper around djvu.decode.Context to handle messages.
    """

    def __init__(self):
        super().__init__()
        self.message_handler = None

    def handle_message(self, message):
        """
        Handles messages from the DjVu decoding context.
        """
        if self.message_handler:
            self.message_handler(message)


class DjVuProcessor:
    """
    Processes DjVu files and converts pages to image buffers.

    see https://raw.githubusercontent.com/jwilk-archive/python-djvulibre/refs/heads/master/examples/djvu2png
    with Copyright © 2010-2021 Jakub Wilk <jwilk@jwilk.net> and GNU General Public License version 2
    """

    def __init__(self, tar: bool = True, verbose: bool = False, debug: bool = False):
        """
        Initializes the DjVuProcessor.

        Args:
            tar(bool,optional): Enable tarball creation (default: True).
            verbose (bool, optional): Enable verbose output (default: False).
            debug (bool, optional): Enable debug logging (default: False).
        """
        self.tar = tar
        self.verbose = verbose
        self.debug = debug
        self.context = DjVuContext()  # delegate context instance
        self.context.message_handler = self.handle_message
        self.cairo_pixel_format = cairo.FORMAT_ARGB32
        self.djvu_pixel_format = djvu.decode.PixelFormatRgbMask(
            0xFF0000, 0xFF00, 0xFF, bpp=32
        )
        self.djvu_pixel_format.rows_top_to_bottom = 1
        self.djvu_pixel_format.y_top_to_bottom = 0

    def create_tarball(
        self, source_dir: str, output_tar: str, include_ext: Optional[List[str]] = None
    ):
        """
        Creates a tar archive from the given source directory, including only specific file types.

        Args:
            source_dir (str): Directory containing files to archive.
            output_tar (str): Path to the output tar file.
            include_ext (Optional[List[str]]): List of file extensions to include.
                - "yaml": Includes metadata files.
                - "png": Includes lossless original images.
                - "jpg": Includes compressed thumbnails.
                Defaults to ["yaml", "png", "jpg"].
        """
        if include_ext is None:
            include_ext = [
                "yaml",
                "png",
                "jpg",
            ]  # yaml metadata, png lossless original, jpg thumbnails
        with tarfile.open(output_tar, "w") as tar:
            for file in os.listdir(source_dir):
                if any(file.lower().endswith(ext) for ext in include_ext):
                    tar.add(os.path.join(source_dir, file), arcname=file)

    def handle_message(self, message):
        if isinstance(message, djvu.decode.ErrorMessage):
            raise Exception(message)

    def save_image_to_png(self, image_job: ImageJob, output_path: str):
        """
        Saves the rendered DjVu page as a PNG file.

        Args:
            image_job (ImageJob): The processed image job containing buffer data
            output_path (str): The path where the PNG file should be saved.
        """
        if not image_job.image or image_job.image.buffer is None:
            raise ValueError("Image buffer not available in ImageJob")

        width, height = image_job.get_size()
        surface = cairo.ImageSurface.create_for_data(
            image_job.image.buffer, cairo.FORMAT_ARGB32, width, height
        )
        surface.write_to_png(output_path)

    def save_as_png(self, image_job: ImageJob, output_dir: str) -> str:
        """
        Save an image job as PNG in the specified directory

        Args:
            image_job: The image job to save
            output_dir: Directory to save to

        Returns:
            Path to the saved PNG file
        """
        output_path = os.path.join(
            output_dir, f"{image_job.prefix}_page_{image_job.page_index:04d}.png"
        )
        image_job.log("save png start")
        # Save PNG
        self.save_image_to_png(image_job, output_path)
        return output_path
        image_job.log("save png done")

    def render_pagejob_to_buffer(self, image_job: ImageJob, mode: int) -> numpy.ndarray:
        """
        Renders a DjVu page job to a color buffer.

        Args:
            image_job (ImageJob): The job containing the pagejob to render
            mode (int): Rendering mode.

        Returns:
            numpy.ndarray: The rendered color buffer.
        """
        if not image_job.pagejob:
            raise ValueError("PageJob not available")

        width, height = image_job.get_size()
        rect = (0, 0, width, height)

        bytes_per_line = cairo.ImageSurface.format_stride_for_width(
            self.cairo_pixel_format, width
        )
        assert bytes_per_line % 4 == 0

        color_buffer = numpy.zeros((height, bytes_per_line // 4), dtype=numpy.uint32)
        image_job.pagejob.render(
            mode,
            rect,
            rect,
            self.djvu_pixel_format,
            row_alignment=bytes_per_line,
            buffer=color_buffer,
        )

        if mode == djvu.decode.RENDER_FOREGROUND:
            mask_buffer = numpy.zeros_like(color_buffer)
            image_job.pagejob.render(
                djvu.decode.RENDER_MASK_ONLY,
                rect,
                rect,
                self.djvu_pixel_format,
                row_alignment=bytes_per_line,
                buffer=mask_buffer,
            )
            color_buffer |= mask_buffer << 24

        color_buffer ^= 0xFF000000  # Apply transparency
        return color_buffer

    def ensure_file_exists(self,path:str):
        if not os.path.isfile(path):
            msg=f"file {path} not found"
            raise ValueError(msg)

    def yield_pages(self, djvu_path: str):
        """
        yield the pages for the given djvu_path
        """
        self.ensure_file_exists(djvu_path)
        document = self.context.new_document(djvu.decode.FileURI(djvu_path))
        document.decoding_job.wait()
        for page in document.pages:
            yield document, page

    def create_image_jobs(self, djvu_path: str, relurl: str) -> List[ImageJob]:
        """
        Create initial image jobs for all pages in the document

        Args:
            djvu_path (str): Path to the DjVu file
            relurl (str): Relative URL

        Returns:
            List[ImageJob]: List of initialized image jobs
        """
        image_jobs = []
        page_index = 0
        for document, page in self.yield_pages(djvu_path):
            page_index += 1
            job = ImageJob(
                djvu_path=djvu_path,
                document=document,
                page=page,
                page_index=page_index,
                relurl=relurl,
                debug=self.debug,
                verbose=self.verbose
            )
            image_jobs.append(job)

        return image_jobs

    def decode_page(self, image_job: ImageJob, wait: bool = True) -> ImageJob:
        """
        Decodes a single page and updates the ImageJob

        Args:
            image_job (ImageJob): The job to process
            wait (bool): Whether to wait for decoding to complete

        Returns:
            ImageJob: Updated image job with pagejob
        """
        try:
            file_size_msg=""
            filepath=image_job.filepath
            # check whether the document is bundled or not
            if image_job.document.type!=2:
                # we need to check the file is external
                self.ensure_file_exists(filepath)
                file_size = os.path.getsize(filepath)
                file_size_msg=(f"{filepath}:{file_size} bytes ")
            image_job.log(f" page.decode {file_size_msg}start")
            pagejob = image_job.page.decode(wait=wait)
            image_job.log(" page.decode done")
            # Update the image job with the decoded page job
            image_job.pagejob = pagejob
        except Exception as e:
            # Store exception but don't raise
            image_job.error = e
        return image_job

    def render_page(
        self, image_job: ImageJob, mode: int = djvu.decode.RENDER_COLOR
    ) -> ImageJob:
        """
        Renders a page and updates the ImageJob with the rendered image

        Args:
            image_job (ImageJob): The job to process
            mode (int): Rendering mode

        Returns:
            ImageJob: Updated image job with rendered image
        """
        try:
            image_job.log(" render start")
            if not image_job.pagejob:
                raise ValueError(f"PageJob not available for page {image_job.page_index}")

            width, height = image_job.get_size()
            color_buffer = self.render_pagejob_to_buffer(image_job, mode)


            image = DjVuImage(
                width=width,
                height=height,
                dpi=image_job.pagejob.dpi,
                page_index=image_job.page_index,
                djvu_path=image_job.relurl,
                path=image_job.filename,
                buffer=color_buffer,
            )

            # Update the image job with the rendered image
            image_job.image = image
            image_job.log(" render done")
        except Exception as e:
            # Store exception but don't raise
            image_job.error = e
        return image_job

    def prepare(self, output_path: str,relurl:str):
        """
        Prepares the output directory and sets up temporary storage if tarball creation is enabled.

        Args:
            output_path (str): The final destination path for output files.
            relurl(str): the relative url to process

        Attributes:
            final_output_path (str): The actual output path where the final files will be stored.
            temp_dir (Optional[str]): A temporary directory for intermediate storage if tarball creation is enabled.
            output_path (str): The working output path (either temporary or final).
            profiler (Profiler): Profiler instance for tracking processing time.
        """
        self.final_output_path = output_path
        if self.tar:
            # Use a temporary directory for intermediate PNG storage
            self.temp_dir = tempfile.mkdtemp()
            self.output_path = self.temp_dir
        else:
            self.output_path = output_path
        self.profiler = Profiler(f"processing {relurl}", profile=self.verbose or self.debug)
        # Prepare output directory if needed
        os.makedirs(self.final_output_path, exist_ok=True)

    def wrap_as_tarball(self, djvu_path: str):
        """
        Wraps processed output files into a tarball

        Args:
            djvu_path (str): The path to the original DjVu file.

        """
        tarball_path = os.path.join(
            self.final_output_path, f"{Path(djvu_path).stem}.tar"
        )
        self.create_tarball(self.output_path, tarball_path)
        shutil.rmtree(self.temp_dir)

    def process(
        self,
        djvu_path: str,
        relurl: str,
        mode: int = djvu.decode.RENDER_COLOR,
        wait: bool = True,
        save_png: bool = False,
        output_path: str = None,
    ) -> Generator[ImageJob, None, None]:
        """
        Converts a DjVu URL to image buffers with sequential decoding and rendering.
        """
        self.prepare(output_path=output_path,relurl=relurl)
        # Step 1: Create image jobs for all pages
        image_jobs = self.create_image_jobs(djvu_path, relurl)
        self.profiler.time(f" create image jobs")

        # Process each page sequentially
        for job in image_jobs:
            # Step 2: Decode the page
            decoded_job = self.decode_page(job, wait)

            # Step 3: Render the page
            rendered_job = self.render_page(decoded_job, mode)
            self.profiler.time(f" process page {rendered_job.page_index:4d}")

            # Step 4: Optionally save to PNG
            if save_png:
                self.save_as_png(rendered_job, self.output_path)

            yield rendered_job

    def process_parallel(
        self,
        djvu_path: str,
        relurl: str,
        mode: int = djvu.decode.RENDER_COLOR,
        wait: bool = True,
        save_png: bool = False,
        output_path: str = None,
    ) -> Generator[ImageJob, None, None]:
        """
        Converts a DjVu URL to image buffers with fully parallel decoding and rendering.

        Args:
            djvu_path (str): Path to the DjVu file.
            relurl (str): Relative URL for referencing the file.
            mode (int, optional): Rendering mode (default: djvu.decode.RENDER_COLOR).
            wait (bool, optional): Whether to wait for processing to complete (default: True).
            save_png (bool, optional): Whether to save output as PNG files (default: False).
            output_path (str, optional): Directory path to save PNG files (default: None).

        Yields:
            Generator[ImageJob, None, None]: A generator yielding image jobs.
        """
        self.prepare(output_path=output_path,relurl=relurl)

        # Step 1: Create image jobs for all pages
        image_jobs = self.create_image_jobs(djvu_path, relurl)
        self.profiler.time(" create image jobs")

        # Step 2: Decode all pages in parallel
        with ThreadPoolExecutor() as executor:
            decode_futures = [
                executor.submit(self.decode_page, job, wait) for job in image_jobs
            ]

        # Step 3 & 4: Render and save all pages in parallel
        max_workers = os.cpu_count() * 4
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            render_futures: List[Future] = []

            # Submit rendering jobs as decoding completes
            for future in decode_futures:
                decoded_job = future.result()
                render_futures.append(
                    executor.submit(self.render_page, decoded_job, mode)
                )

            # Process rendered jobs as they become available
            for future in render_futures:
                rendered_job = future.result()
                self.profiler.time(f" process page {rendered_job.page_index:4d}")

                # Optionally save to PNG in parallel (submit to executor)
                if save_png:
                    executor.submit(self.save_as_png, rendered_job, self.output_path)

                yield rendered_job



"""
Created on 2025-02-25

@author: wf
"""

import argparse
import logging
import os
import time
import traceback
from dataclasses import asdict

from ngwidgets.profiler import Profiler
from tqdm import tqdm

from genwiki.djvu_core import DjVu, DjVuFile, DjVuPage
from genwiki.djvu_manager import DjVuManager
from genwiki.djvu_processor import DjVuProcessor, ImageJob
from pip._vendor import DEBUNDLED


class DjVuCmd:
    """
    command line handling for djvu processing/converting
    """

    # @FIXME - remove hard coded default_base_path
    default_base_path = os.getenv(
        "GENWIKI_PATH", "/Users/wf/hd/wf-fur.bitplan.com/genwiki"
    )

    def __init__(self, args: argparse.Namespace):
        self.args = args
        self.errors = []

    @classmethod
    def get_argparser(cls) -> argparse.ArgumentParser:
        """
        Get the argument parser for the DjVu command
        """
        output_path = os.path.join(cls.default_base_path, "djvu_images")
        parser = argparse.ArgumentParser(description="Process DjVu files")  #
        parser.add_argument(
            "--base-path",
            default=cls.default_base_path,
            help="Base path for DjVu files",
        )
        parser.add_argument(
            "--command",
            choices=["catalog", "convert","thumbnails","dbupdate"],
            required=True,
            help="Command to execute",
        )
        parser.add_argument(
            "-d",
            "--debug",
            action="store_true",
            default=False,
            help="Enable debugging",
        )
        parser.add_argument(
            "--db-path", default="/tmp/genwiki_djvu.db", help="Path to the database"
        )
        parser.add_argument(
            "-f",
            "--force",
            action="store_true",
            default=False,
            help="Force recreation",
        )
        parser.add_argument(
            "--limit",
            type=int,
            default=10000000,
            help="Maximum number of pages to process",
        )
        parser.add_argument(
            "--output-path", default=output_path, help="Path for PNG files"
        )
        parser.add_argument(
            "--serial", action="store_true", help="Use serial processing - parallel is default"
        )
        parser.add_argument(
            "--sort",
            choices=["asc", "desc"],
            default="asc",
            help="Sort by page count (asc=smallest first)",
        )
        parser.add_argument(
            "-v",
            "--verbose",
            action="store_true",
            default=False,
            help="Enable debugging",
        )
        parser.add_argument(
            "--url", help="Process a single DjVu file (only valid in convert mode)"
        )

        return parser

    def handle_args(self):
        """
        handle the command line arguments
        """
        self.dvm = DjVuManager(db_path=self.args.db_path)
        self.dproc = DjVuProcessor(debug=self.args.debug, verbose=self.args.verbose)
        self.profiler = Profiler(self.args.command)
        self.profiler.start()
        if self.args.command == "catalog":
            self.catalog_djvu()
        elif self.args.command == "convert":
            self.convert_djvu()
        elif self.args.command == "thumbnails":
            self.generate_thumbnails()
        elif self.args.command == "dbupdate":
            self.update_database()

    def add_page(self, page_lod, path: str, page_index: int, page):
        """
        Adds a DjVuPage to the given list of pages.

        Args:
            page_lod (list): List to which the page dictionary is appended.
            path (str): Path to the DjVu file.
            page_index (int): Index of the page.
            page: Page object containing metadata.

        Returns:
            DjVuPage: The created DjVuPage instance.
        """
        try:
            filename = page.file.name
            if "gesperrtes" in filename:
                filename = "?"
                valid = False
            else:
                valid = True
        except Exception as _ex:
            filename = "?"
            valid = False
        dpage = DjVuPage(
            # height=page.height,
            # width=page.width,
            # dpi=page.dpi,
            path=filename,
            page_index=page_index,
            valid=valid,
            djvu_path=path,
        )
        row = asdict(dpage)
        page_lod.append(row)
        return dpage

    def report_errors(self):
        """
        Reports errors collected during processing.

        - ✅ Shows ✅ (check mark) if no errors.
        - ❌ Shows ❌ (cross mark) with error count if errors occurred.
        - 📝 Lists errors if `args.debug` is enabled.
        - 📜 Shows stack traces if `args.verbose` is enabled.
        """
        if not self.errors:
            msg = " ✅ Ok"
        else:
            msg = f" ❌ {len(self.errors)} errors"
        self.profiler.time(msg)

        if self.args.debug:
            for i, error in enumerate(self.errors, 1):
                print(f"📝 {i}. {error}")

            if self.args.verbose:
                for error in self.errors:
                    print("📜", traceback.format_exc())

    def catalog_djvu(self):
        """
        First pass: Catalog DjVu files into the database
        """
        bootstrap_dvm = DjVuManager()
        lod = bootstrap_dvm.query("all_djvu")
        total = 0
        start_time = time.time()
        djvu_lod = []
        page_lod = []
        for index, r in enumerate(lod, start=1):
            path = r.get("path").replace("./", "/")
            djvu_path = self.args.base_path + path
            if not djvu_path:
                self.errors.append(Exception(f"missing {djvu_path}"))
                continue
            page_index = 0
            for document, page in self.dproc.yield_pages(djvu_path):
                page_count = len(document.pages)
                page_index += 1
                _dpage = self.add_page(page_lod, path, page_index, page)
                bundled=document.type==2
                # if debug:
                #    print(f"    {page_index:4d}/{page_count:4d}:{filename}")
            djvu = DjVu(
                path=path,
                page_count=page_count,
                bundled=bundled
            )
            djvu_row = asdict(djvu)
            djvu_lod.append(djvu_row)
            total += page_index
            if total > self.args.limit:
                break
            elapsed = time.time() - start_time
            pages_per_sec = total / elapsed if elapsed > 0 else 0
            print(
                f"{index:4d} {page_count:4d} {total:7d} {pages_per_sec:7.0f} pages/s: {path}"
            )
        self.dvm.store(lod=page_lod, entity_name="Page", primary_key="page_key")
        self.dvm.store(lod=djvu_lod, entity_name="DjVu", primary_key="path")
        self.report_errors()

    def get_djvu_files(self):
        # Handle single-file mode
        if self.args.url:
            djvu_files = [self.args.url]
        else:
            lod = self.dvm.query("all_djvu")
            djvu_files = [r.get("path").replace("./", "/") for r in lod]
        return djvu_files

    def convert_djvu(self):
        """
        Second pass: Convert DjVu files to PNG using the database
        """
        djvu_files=self.get_djvu_files()
        # select the process function parallel or serial
        process_func=self.dproc.process if self.args.serial else self.dproc.process_parallel
        with tqdm(
            total=len(djvu_files), desc="Converting DjVu to PNG", unit="file"
        ) as pbar:
            for path in djvu_files:
                try:
                    djvu_path = self.args.base_path + path
                    djvu_file = None
                    prefix = ImageJob.get_prefix(path)
                    tar_file = os.path.join(self.args.output_path, prefix + ".tar")
                    if os.path.isfile(tar_file) and not self.args.force:
                        continue
                    for image_job in process_func(
                        djvu_path,
                        relurl=path,
                        save_png=True,
                        output_path=self.args.output_path,
                    ):
                        # collect upstream errors
                        if hasattr(image_job, 'error') and image_job.error:
                            self.errors.append(image_job.error)
                            continue
                        if djvu_file is None:
                            page_count = len(image_job.document.pages)
                            djvu_file = DjVuFile(path=path, page_count=page_count)
                        image = image_job.image
                        djvu_page = DjVuPage(
                            path=image.path,
                            page_index=image.page_index,
                            valid=image.valid,
                            width=image.width,
                            height=image.height,
                            dpi=image.dpi,
                            djvu_path=image.djvu_path,
                        )
                        djvu_file.pages.append(djvu_page)
                        prefix = image_job.prefix
                        pass
                    yaml_file = os.path.join(self.dproc.output_path, prefix + ".yaml")
                    djvu_file.save_to_yaml_file(yaml_file)

                    # Ensure tarball is created after YAML is saved
                    if self.dproc.tar:
                        self.dproc.wrap_as_tarball(djvu_path)
                except BaseException as e:
                    self.errors.append(e)
                finally:
                    error_count=len(self.errors)
                    status_msg="✅" if error_count==0 else f"❌ {error_count}"
                    pbar.set_postfix_str(status_msg)
                    pbar.update(1)
        self.report_errors()

    def generate_thumbnails(self):
        """
        Generates thumbnails for all DjVu files.
        """
        self.errors.append(Exception("generate thumbnails not implemented yet"))
        self.report_errors()

    def update_database(self):
        """
        Updates the DjVu database.
        """
        djvu_files=self.get_djvu_files()
        with tqdm(
            total=len(djvu_files), desc="Converting DjVu to PNG", unit="file"
        ) as pbar:
            for path in djvu_files:
                try:
                    djvu_path = self.args.base_path + path
                    prefix = ImageJob.get_prefix(path)
                    tar_file = os.path.join(self.args.output_path, prefix + ".tar")
                    if not os.path.isfile(tar_file):
                        raise Exception(f"tar file for {path} missing")
                except BaseException as e:
                    self.errors.append(e)
                finally:
                    error_count=len(self.errors)
                    status_msg="✅" if error_count==0 else f"❌ {error_count}"
                    pbar.set_postfix_str(status_msg)
                    pbar.update(1)
        self.report_errors()


def main():
    """
    Command-line interface for processing DjVu files and updating the database
    """
    parser = DjVuCmd.get_argparser()
    args = parser.parse_args()
    cmd = DjVuCmd(args)
    cmd.handle_args()


if __name__ == "__main__":
    # Setup logging
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    )
    main()



"""
Created on 24.10.2024

@author: wf
"""

import argparse
import sys
from typing import Any, Dict, List

import requests
from tqdm import tqdm

# Color definitions
BLUE = "\033[0;34m"
RED = "\033[0;31m"
GREEN = "\033[0;32m"
ENDC = "\033[0m"


class Playground:
    """
    a class to create wiki playgrounds
    """

    def __init__(self, base_port: int = 9100):
        self.base_port = base_port
        self.step = 2  # Port step (for mw_port and sql_port)

    def color_msg(self, color: str, msg: str) -> None:
        """Displays colored messages."""
        print(f"{color}{msg}{ENDC}")

    def godsList(self) -> List[Dict[str, Any]]:
        """ """
        return [
            {
                "name": "Aglaea",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/20140416%20corfu218.JPG",
            },
            {
                "name": "Amicitia",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/Amicitia%20DenHaag%20Friedenspalast.JPG",
            },
            {
                "name": "Anteros",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/Eros-piccadilly-circus.jpg",
            },
            {
                "name": "Ares",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/%CE%9F%20%CE%86%CF%81%CE%B7%CF%82%20%28Borghese-%CE%9B%CE%BF%CF%8D%CE%B2%CF%81%CE%BF%CF%85%29.jpg",
            },
            {
                "name": "Aristaeus",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/Aristaeus%20Bosio%20Louvre%20LL51.jpg",
            },
            {
                "name": "Artemis",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/The%20Artemis%20of%20Ephesus.jpg",
            },
            {
                "name": "Bia",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/Henry%20Fuseli%20-%20Hephaestus%2C%20Bia%20and%20Crato%20Securing%20Prometheus%20on%20Mount%20Caucasus%20-%20Google%20Art%20Project.jpg",
            },
            {
                "name": "Britomartis",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/Email%20Limoges%20Spiegelr%C3%BCckseite%20Minos%20und%20Britomaris%20makffm%20WMH8.jpg",
            },
            {
                "name": "Carpo",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/Carpo-hora%2C%20I%20secolo%20dc%2C%20%28Uffizi%29%2003.JPG",
            },
            {
                "name": "Charon",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/Sabouroff%20Painter%20ARV%20846%20196%20Hermes%20leading%20a%20deceased%20to%20Charon%20%2802%29.jpg",
            },
            {
                "name": "Deimos",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/Deimos-e-phoboslllll.jpg",
            },
            {
                "name": "Demeter",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/%28Venice%29%20Demeter%20in%20the%20Museo%20archeologico%20nazionale.jpg",
            },
            {
                "name": "Despoina",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/Lycosoura-group%20%28cropped%29.jpg",
            },
            {
                "name": "Eileithyia",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/Amphora%20birth%20Athena%20Louvre%20F32.jpg",
            },
            {
                "name": "Eirene",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/Eirene%20von%20Knaus.jpg",
            },
            {
                "name": "Enyo",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/Image%20of%20Enyo%20the%20Goddess-%202014-04-16%2007-17.jpg",
            },
            {
                "name": "Eos",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/Eos%20utgjutande%20morgondaggen%2C%20Nordisk%20familjebok.png",
            },
            {
                "name": "Eris",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/Eris%20Antikensammlung%20Berlin%20F1775.jpg",
            },
            {
                "name": "Eunomia",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/Horen%20Meyers.jpg",
            },
            {
                "name": "Euphrosyne",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/Euphrosyne%20statue%20-%20Achilleion.jpg",
            },
            {
                "name": "Hades",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/Detail%20of%20Pluto-Serapis%2C%20Statue%20group%20of%20Persephone%20%28as%20Isis%29%20and%20Pluto%20%28as%20Serapis%29%2C%20from%20the%20Sanctuary%20of%20the%20Egyptian%20Gods%20at%20Gortyna%2C%20mid-2nd%20century%20AD%2C%20Heraklion%20Archaeological%20Museum%20%2830305313721%29.jpg",
            },
            {
                "name": "Hebe",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/Canova-Hebe%2030%20degree%20view.jpg",
            },
            {
                "name": "Helios",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/Architrave%20with%20sculpted%20metope%20showing%20sun%20god%20Helios%20in%20a%20quadriga%3B%20from%20temple%20of%20Athena%20at%20Troy%2C%20ca%20300-280%20BCE%3B%20Altes%20Museum%2C%20Berlin%20%2825308440197%29%20%28cropped%29%201.jpg",
            },
            {
                "name": "Hemera",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/Aphrodisias%20Museum%20Hemera%20or%20Day%204627.jpg",
            },
            {
                "name": "Hephaestus",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/Hephaistos%20Thetis%20at%20Kylix%20by%20the%20Foundry%20Painter%20Antikensammlung%20Berlin%20F2294.jpg",
            },
            {
                "name": "Hera",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/Hera%20Campana%20Louvre%20Ma2283.jpg",
            },
            {
                "name": "Hermaphroditus",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/Ermafrodito%20-%20sec.%20III%20a.C.%20-%20da%20Pergamo.%20Istanbul.%20Museo%20archeol..jpg",
            },
            {
                "name": "Hestia",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/Hestia%20-%20Wellesley%20College%20-%20DSC09634.JPG",
            },
            {
                "name": "Himeros",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/Kantharos64.10.jpg",
            },
            {
                "name": "Hypnos",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/Waterhouse-sleep%20and%20his%20half-brother%20death-1874.jpg",
            },
            {
                "name": "Iris",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/Iris%20Louvre%20L43%20n2.jpg",
            },
            {
                "name": "Kratos",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/Kratos%20by%20John%20Flaxman.jpg",
            },
            {
                "name": "Momus",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/Nantes%20-%20Graslin%20int%2001.jpg",
            },
            {
                "name": "Nike",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/Nik%C3%A9%20%C3%89ph%C3%A8se.jpg",
            },
            {
                "name": "Panacea",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/Panacea.jpg",
            },
            {
                "name": "Phobos",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/Gigantomachy%20Staatliche%20Antikensammlungen%201553.jpg",
            },
            {
                "name": "Phosphorus",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/Evelyn%20de%20Morgan%20-%20Phosphorus%20and%20Hesperus%2C%20%281881%29.jpg",
            },
            {
                "name": "Plutus",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/Eirene%20Ploutos%20Glyptothek%20Munich%20219%20n4.jpg",
            },
            {
                "name": "Poseidon",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/Poseidon%20sculpture%20Copenhagen%202005.jpg",
            },
            {
                "name": "Telesphorus",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/Statue%20of%20Telesphorus%2C%20Greece%2C%20500-200%20BCE%20Wellcome%20L0058862.jpg",
            },
            {
                "name": "Thalia",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/Thalia%20at%20Corfu%201.jpg",
            },
            {
                "name": "Thallo",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/The%20Spring%20by%20Franz%20Xaver%20Winterhalter.jpg",
            },
            {
                "name": "Thanatos",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/Euphronios%20krater%20side%20A%20MET%20L.2006.10.jpg",
            },
            {
                "name": "Zeus",
                "img": "http://commons.wikimedia.org/wiki/Special:FilePath/Jupiter%20J1a.jpg",
            },
        ]

    def generate_header(self, image_url: str) -> str:
        """
        Generates the HTML header section with the logo on the left and the title aligned to the right on the same line.
        """
        return f"""
    <!DOCTYPE html>
    <html>
    <head>
        <title>CompGen - Semantic MediaWiki Spielwiesen</title>
        <style>
            body {{ font-family: Arial, sans-serif; background-color: #f5f5f5; padding: 0; margin: 0; }}
            .header {{
                background-color: #2d4899;
                padding: 10px 20px;
                display: flex;
                align-items: center;
                justify-content: space-between;
            }}
            .header img {{ width: 100px; }}
            .header h1 {{ color: white; font-size: 24px; margin: 0; }}
            .container {{ max-width: 1200px; margin: 20px auto; padding: 10px; }}
            .grid-container {{
                display: grid;
                grid-template-columns: repeat(auto-fill, minmax(100px, 1fr));
                gap: 10px;
                justify-items: center;
            }}
            .grid-item {{
                text-align: center;
            }}
            .grid-item img {{ width: 100px; }}
            a {{ color: #2d4899; text-decoration: none; }}
            a:hover {{ text-decoration: underline; }}
        </style>
    </head>
    <body>
        <div class="header">
            <img src="{image_url}" alt="CompGen Logo">
            <h1>Semantic MediaWiki Spielwiesen</h1>
        </div>
        <div class="container">
        """

    def get_thumbnail_url(self, img_url: str, size: int = 300) -> str:
        """
        Generates the thumbnail URL from the original image URL by following the redirect from Special:FilePath.

        Args:
            img_url (str): The original image URL or the redirected URL.
            size (int, optional): The desired size of the thumbnail. Defaults to 300px.

        Returns:
            str: The modified URL pointing to the thumbnail.
        """
        # Follow the redirect
        response = requests.get(img_url)
        final_url = response.url  # Get the final URL after redirection

        # Generate the thumbnail URL
        thumb_url = (
            final_url.replace("/commons/", f"/commons/thumb/")
            + f"/{size}px-"
            + final_url.split("/")[-1]
        )
        return thumb_url

    def generate_table(self) -> str:
        """
        Generates the HTML grid with the list of Greek gods and their images, using thumbnails, and limits to 9 images per row.
        """
        gods = self.godsList()
        table_content = '<div class="grid-container">\n'
        count = 0
        for god in tqdm(gods, desc="Generating thumbnails"):
            name = god["name"]
            img = god["img"]
            # Use the separate function to get the thumbnail URL
            thumb_img = self.get_thumbnail_url(img, 150)
            table_content += f'<div class="grid-item"><a href="/{name}">{name}</a><br><img src="{thumb_img}" alt="{name}"></div>\n'
            count += 1
            # Break after 9 images
            if count % 9 == 0:
                table_content += '</div><div class="grid-container">\n'
        table_content += "</div>\n"
        return table_content

    def generate_footer(self) -> str:
        """
        Generates the HTML footer section.
        """
        return """
    </body>
    </html>
    """

    def generate_index(
        self,
        file_path: str = "/tmp/index.html",
        logo_url: str = "https://www.compgen.de/wp-content/uploads/2019/01/CG-Logo02-340_156px-200x92.png",
    ) -> None:
        """
        Combines the header, table, and footer to generate the full HTML file (index.html).
        """
        header = self.generate_header(logo_url)
        table = self.generate_table()
        footer = self.generate_footer()

        with open(file_path, "w") as file:
            file.write(header + table + footer)

    def generate_apache_conf(self, output_path: str = "/tmp/playground.conf") -> None:
        """
        Generates the Apache configuration file for all wikis
        """
        conf_content = """#
# Proxy configuration for Wiki Playgrounds
# generated by playground.py
# https://github.com/WolfgangFahl/genwiki2024/blob/main/genwiki/playground.py
#
<VirtualHost *:80>
    ServerName playground-mw.bitplan.com
    ProxyPreserveHost On
    """
        for idx, god in enumerate(self.godsList()):
            mw_port = self.base_port + (idx * self.step)
            conf_content += f"""    ProxyPass /{god['name']} http://localhost:{mw_port}
    ProxyPassReverse /{god['name']} http://localhost:{mw_port}
    """
        conf_content += """    CustomLog ${APACHE_LOG_DIR}/playground_access.log combined
    ErrorLog ${APACHE_LOG_DIR}/playground_error.log
</VirtualHost>
    """
        with open(output_path, "w") as f:
            f.write(conf_content)
        return output_path

    def generate_setup_script(self, output_path: str = "/tmp/setup_wikis.sh") -> None:
        """
        Generates a bash script to create and configure all wikis
        """
        script_content = """#!/bin/bash
# Setup script for Wiki Playgrounds
# generated by playground.py
# https://github.com/WolfgangFahl/genwiki2024/blob/main/genwiki/playground.py

set -e  # Exit on error
"""
        # First create all wikis
        for idx, god in enumerate(self.godsList()):
            mw_port = self.base_port + (idx * self.step)
            sql_port = mw_port + 1
            name = god["name"]
            script_content += f"""
echo "Setting up {name} wiki..."
profiwiki -rp -fu -cn {name} -bp {mw_port} -sp {sql_port} --all -f
"""
        with open(output_path, "w") as f:
            f.write(script_content)
        return output_path


def main():
    parser = argparse.ArgumentParser(description="Wiki Playground Generator")
    parser.add_argument("--apache", action="store_true", help="Generate apache config")
    parser.add_argument("--index", action="store_true", help="Generate index.html")
    parser.add_argument("--list", action="store_true", help="List the gods")
    parser.add_argument(
        "--setup", action="store_true", help="Generate the setup script"
    )
    args = parser.parse_args()

    pg = Playground()
    handled = False
    if args.index:
        pg.generate_index()
        pg.color_msg(GREEN, "✅ index.html generated.")
        handled = True
    if args.apache:
        conf = pg.generate_apache_conf()
        pg.color_msg(GREEN, f"✅ {conf} generated.")
        handled = True
    if args.setup:
        setup = pg.generate_setup_script()
        pg.color_msg(GREEN, f"✅ {setup} generated.")
        handled = True
    if args.list:
        gods = pg.godsList()
        for god in gods:
            print(god["name"])
        handled = True
    if not handled:
        parser.print_help()  # Replaces usage
        sys.exit(1)


if __name__ == "__main__":
    main()



"""
Created on 25.08.2024

@author: wf
"""

import logging

from geopy.exc import GeocoderQueryError, GeocoderTimedOut, GeocoderUnavailable
from geopy.geocoders import Nominatim


class NominatimWrapper:
    """
    Nominatim Wrapper to search for locations and retrieve Wikidata IDs
    """

    def __init__(self, user_agent: str = "AppUsingNominatim"):
        """
        Initialize the NominatimWrapper

        Args:
            user_agent (str): The user agent to use for the geolocator
        """
        self.geolocator = Nominatim(user_agent=user_agent)
        logging.getLogger("geopy").setLevel(logging.ERROR)

    def lookup_wikidata_id(self, location_text: str, max_retries: int = 3):
        """
        Lookup the Wikidata Identifier for the given location text

        Args:
            location_text (str): The location text to search for
            max_retries (int): Maximum number of retries in case of timeout or unavailability

        Returns:
            str: The Wikidata Q identifier most fitting the given location text, or None if not found
        """
        for attempt in range(max_retries):
            try:
                location = self.geolocator.geocode(
                    location_text, exactly_one=True, extratags=True
                )

                if location:
                    extratags = location.raw.get("extratags", {})
                    if extratags and "wikidata" in extratags:
                        return extratags["wikidata"]
                return None
            except (GeocoderTimedOut, GeocoderUnavailable):
                if attempt == max_retries - 1:
                    logging.error(f"Failed to geocode after {max_retries} attempts")
                else:
                    logging.warning(
                        f"Geocoding attempt {attempt + 1} failed, retrying..."
                    )
            except GeocoderQueryError:
                logging.warning("Geocoding failed")
        return None



"""
Created on 2024-08-15

@author: wf
"""

import logging
import os
from typing import Any, Dict, Generator, List, Tuple

from wikibot3rd.wikipush import WikiPush


class Wiki:
    """
    Access to a Mediawiki.

    Attributes:
        wiki_id (str): The ID of the wiki to access.
        wiki_backup_dir (str): The directory where the backup files are stored.
        debug (bool): Indicates if debugging is enabled.
    """

    def __init__(self, wiki_id: str, backup_dir: str = None, debug: bool = False):
        """
        Initialize the Wiki object.

        Args:
            wiki_id (str): The ID of the wiki to access.
            backup_dir(str): the backup directory to use
            debug (bool): If True, enable debug logging.
        """
        self.wiki_id = wiki_id
        self.wiki_push = WikiPush(fromWikiId=wiki_id, debug=debug)
        self.debug = debug
        if backup_dir is None:
            backup_dir = os.path.expanduser(f"~/wikibackup/{self.wiki_id}")
        self.wiki_backup_dir = backup_dir

        # Set up logging
        if self.debug:
            logging.basicConfig(level=logging.DEBUG)
        else:
            logging.basicConfig(level=logging.INFO)

    def log(self, message: str, level: int = logging.INFO):
        """
        Log a message based on the debug setting.

        Args:
            message (str): The message to log.
            level (int): The logging level, default is INFO.
        """
        if self.debug:
            logging.log(level, message)
        else:
            logging.log(logging.DEBUG, message)

    def query_as_dict_of_dicts(self, ask_query: str) -> Dict[str, Dict[str, Any]]:
        """
        run the given SMW ask query against my wiki
        """
        qdict = self.wiki_push.queryPages(askQuery=ask_query)
        return qdict

    def query_as_list_of_dicts(self, ask_query: str) -> List[Dict[str, Any]]:
        qdict = self.query_as_dict_of_dicts(ask_query)
        qlod = list(qdict.values())
        return qlod

    def backup(self, ask_query: str):
        """
        backup the pages for the given query
        """
        page_lod = self.wiki_push.queryPages(ask_query)
        page_titles = list(page_lod.keys())
        self.wiki_push.backup(pageTitles=page_titles, backupPath=self.wiki_backup_dir)

    def get_all_wiki_files(self) -> Generator[Tuple[str, str], None, None]:
        """
        Generator that yields tuples of (file_path, page_name)
        for all .wiki files in the backup directory.
        """
        if not os.path.isdir(self.wiki_backup_dir):
            err_msg = f"Directory {self.wiki_backup_dir} does not exist."
            self.log(err_msg, logging.ERROR)
            raise ValueError(err_msg)

        for root, _, files in os.walk(self.wiki_backup_dir):
            for file_name in files:
                if file_name.endswith(".wiki"):
                    file_path = os.path.join(root, file_name)
                    page_name = file_name[:-5]  # Remove .wiki extension
                    relative_path = os.path.relpath(file_path, self.wiki_backup_dir)
                    page_name = os.path.splitext(relative_path)[
                        0
                    ]  # Remove extension and use relative path
                    self.log(f"Processing file: {file_path}", logging.DEBUG)
                    yield file_path, page_name

    def get_all_content(self) -> Dict[str, str]:
        """
        Gather the content of all .wiki files in the backup directory.

        Returns:
            Dict[str, str]: A dictionary where keys are page names and values are the content of the .wiki files.
        """
        content_dict = {}
        for file_path, page_name in self.get_all_wiki_files():
            content_dict[page_name] = self.read_file_content(file_path)
        return content_dict

    def read_file_content(self, file_path: str) -> str:
        """
        Read the content of a .wiki file.

        Args:
            file_path (str): The path to the .wiki file.

        Returns:
            str: The content of the file.
        """
        self.log(f"Attempting to read file: {file_path}", logging.DEBUG)
        with open(file_path, "r", encoding="utf-8") as file:
            content = file.read()
            self.log(f"Read content from {file_path}", logging.DEBUG)
            return content



"""
Created on 2024-08-26

@author: wf
"""

import logging
import os
from typing import Any, Dict, List

import pyarrow.parquet as pq
from lodstorage.schema import Schema, SchemaManager
from lodstorage.sql import SQLDB, EntityInfo


class Parquet:
    """
    A class to handle Parquet file operations and SQLite conversions.

    This class provides functionality to read Parquet files from a directory
    and convert their contents to an SQLite database using EntityInfo.

    Attributes:
        debug (bool): If True, debug information will be logged.
    """

    def __init__(self, debug: bool = False):
        """
        Initialize the Parquet handler.

        Args:
            debug (bool): If True, enables debug output. Defaults to False.
        """
        self.debug = debug

    def log(self, msg: str) -> None:
        """
        Log a message if debug is True.

        Args:
            msg (str): The message to log.
        """
        if self.debug:
            logging.debug(msg)

    def read_parquet_files(self, directory: str) -> Dict[str, List[Dict[str, Any]]]:
        """
        Read all Parquet files in the given directory and return their contents as a dictionary of table names to rows.

        Args:
            directory (str): The path to the directory containing Parquet files.

        Returns:
            Dict[str, List[Dict[str, Any]]]: A dictionary where keys are table names (derived from file names)
                                             and values are lists of dictionaries representing the rows.

        Raises:
            FileNotFoundError: If the specified directory does not exist.
        """
        if not os.path.exists(directory):
            raise FileNotFoundError(f"Directory not found: {directory}")

        parquet_files = [f for f in os.listdir(directory) if f.endswith(".parquet")]
        tables_data = {}

        for parquet_file in parquet_files:
            file_path = os.path.join(directory, parquet_file)
            table_name = os.path.splitext(parquet_file)[0]  # Remove .parquet extension
            try:
                table = pq.read_table(file_path)
                rows = table.to_pylist()
                # Add source field to each row
                for row in rows:
                    row["source"] = parquet_file
                tables_data[table_name] = rows
                log_msg = f"Read {len(rows)} rows from {parquet_file}"
                self.log(log_msg)
            except Exception as e:
                log_msg = f"Failed to read {parquet_file}: {e}"
                self.log(log_msg)

        return tables_data

    def convert_parquet_to_sqlite(
        self,
        parquet_data: Dict[str, List[Dict[str, Any]]],
        db: SQLDB,
        table_name: str = None,
        column_mapping: Dict[str, str] = None,
    ):
        """
        Converts Parquet data to SQLite tables and creates a combined view.

        Args:
            parquet_data (Dict[str, List[Dict[str, Any]]]): A dictionary where keys are table names and values are lists
                of dictionaries representing rows of data. Each dictionary in the list represents a row, with keys as
                column names and values as column values.
            db (SQLDB): An instance of the SQLDB class where the tables will be created.
            table_name (str, optional): The name of the combined view that will be created. If not provided,
                defaults to "combined_view".
            column_mapping (Dict[str, str], optional): A dictionary mapping original column names to new column names.
                If provided, the columns in the Parquet data will be renamed accordingly.

        Raises:
            ValueError: If `parquet_data` is empty.

        Returns:
            None

        Example:
            parquet_data = {
                "table1": [{"col1": "value1", "col2": "value2"}],
                "table2": [{"col1": "value1", "col2": "value2"}]
            }
            db = SQLDB()
            convert_parquet_to_sqlite(parquet_data, db, table_name="my_view", column_mapping={"col1": "new_col1"})

        This method performs the following steps:
            1. Converts each table in `parquet_data` into a corresponding SQLite table.
            2. Applies `column_mapping` to rename columns, if provided.
            3. Creates a combined view in the SQLite database that includes all the tables.
            4. Logs the progress of adding rows to each SQLite table and creating the view.
        """
        if not parquet_data:
            raise ValueError("No data to convert to SQLite")

        table_list = []

        for original_table_name, rows in parquet_data.items():
            if column_mapping:
                rows = self._apply_column_mapping(rows, column_mapping)

            entityInfo = EntityInfo(rows, original_table_name, debug=self.debug)
            db.createTable4EntityInfo(entityInfo)
            db.store(rows, entityInfo)

            table = {
                "name": original_table_name,
                "columns": [
                    {"name": col, "type": str(type_)}
                    for col, type_ in entityInfo.typeMap.items()
                ],
            }
            table_list.append(table)

            self.log(f"Added {len(rows)} rows to SQLite table '{original_table_name}'")

        # Create a view that combines all tables
        view_name = table_name or "combined_view"
        view_ddl = Schema.getGeneralViewDDL(table_list, view_name, debug=self.debug)
        db.execute(view_ddl)
        self.log(f"Created view '{view_name}' combining all tables")

    def _apply_column_mapping(
        self, rows: List[Dict[str, Any]], column_mapping: Dict[str, str]
    ) -> List[Dict[str, Any]]:
        """
        Apply the column mapping to the rows.

        Args:
            rows (List[Dict[str, Any]]): The original rows of data.
            column_mapping (Dict[str, str]): Mapping of Parquet column names to SQLite column names.

        Returns:
            List[Dict[str, Any]]: The rows with updated column names.
        """
        return [{column_mapping.get(k, k): v for k, v in row.items()} for row in rows]



"""
Created on 2025-02-25

@author: wf
"""

import os
from dataclasses import dataclass, field
from typing import List, Optional

import numpy
from ngwidgets.yamlable import lod_storable


@lod_storable
class DjVuPage:
    """Represents a single djvu page"""

    path: str
    page_index: int
    valid: bool = False
    width: Optional[int] = None
    height: Optional[int] = None
    dpi: Optional[int] = None
    djvu_path: Optional[str] = None
    page_key: Optional[str] = None

    def __post_init__(self):
        """Post-initialization logic for DjVuPage."""
        if self.page_key is None:
            # we expect no more than 9999 pages per document in the genwiki context that is proven
            self.page_key = f"{self.djvu_path}#{self.page_index:04d}"
        pass

    @property
    def png_file(self) -> str:
        """
        Returns the PNG file name derived from the DjVu file path and page index.
        """
        prefix = os.path.splitext(os.path.basename(self.djvu_path))[0]
        png_file = f"{prefix}_page_{self.page_index:04d}.png"
        return png_file


@dataclass
class DjVu:
    """Represents a DjVu main file e.g. bundled or indexed"""

    path: str
    page_count: int
    bundled: bool=False
    dir_pages: Optional[int] = None


@lod_storable
class DjVuFile(DjVu):
    """Represents a DjVu main file e.g. bundled or indexed"""

    pages: List[DjVuPage] = field(default_factory=list)

    def get_page_by_page_index(self, page_index: int) -> Optional[DjVuPage]:
        """
        Retrieve a page by its page index.

        Args:
            page_index (int): The index of the page to retrieve.

        Returns:
            Optional[DjVuPage]: The requested DjVuPage if found, otherwise None.
        """
        for page in self.pages:
            if page.page_index == page_index:
                return page
        return None


@dataclass
class DjVuImage(DjVuPage):
    buffer: Optional[numpy.ndarray] = None



"""
Created on 15.09.2024

@author: wf
"""

import os


class GenWikiPaths:
    @classmethod
    def get_examples_path(cls) -> str:
        # the root directory (default: examples)
        path = os.path.join(os.path.dirname(__file__), "../genwiki_examples")
        path = os.path.abspath(path)
        return path



"""
Created on 2024-08-19

Basetests for Genealogy Wiki
@author: wf
"""

import os

from ngwidgets.basetest import Basetest
from wikibot3rd.wikiuser import WikiUser


class GenealogyBasetest(Basetest):

    def setUp(self, debug=False, profile=True):
        Basetest.setUp(self, debug=debug, profile=profile)

    def get_wiki_user(self, wikiId="or", save=False) -> WikiUser:
        """
        get wiki users for given wikiId
        """
        iniFile = WikiUser.iniFilePath(wikiId)
        wikiUser = None
        if not os.path.isfile(iniFile):
            wikiDict = None
            if wikiId == "cr":
                wikiDict = {
                    "wikiId": wikiId,
                    "email": "noreply@nouser.com",
                    "url": "http://cr.bitplan.com",
                    "scriptPath": "",
                    "version": "MediaWiki 1.35.5",
                }
            elif wikiId == "genealogy":
                wikiDict = {
                    "wikiId": wikiId,
                    "email": "noreply@nouser.com",
                    "url": "https://wiki.genealogy.net",
                    "scriptPath": "/",
                    "version": "MediaWiki 1.35.11",
                }
            elif wikiId == "gensmw":
                wikiDict = {
                    "wikiId": wikiId,
                    "email": "noreply@nouser.com",
                    "url": "https://gensmw.bitplan.com/",
                    "scriptPath": "/",
                    "version": "MediaWiki 1.39.8",
                }
            elif wikiId == "smwcopy":
                wikiDict = {
                    "wikiId": wikiId,
                    "email": "noreply@nouser.com",
                    "url": "https://smw.bitplan.com/",
                    "scriptPath": "",
                    "version": "MediaWiki 1.35.5",
                }
            if wikiDict is None:
                raise Exception(f"wikiId {wikiId} is not known")
            else:
                wikiUser = WikiUser.ofDict(wikiDict, lenient=True)
                if save:
                    wikiUser.save()
        else:
            wikiUser = WikiUser.ofWikiId(wikiId, lenient=True)
        return wikiUser



"""
Created on 25.02.2025

@author: wf
"""

import os
import time

import requests


class Download:
    @classmethod
    def download(cls, url: str, target_path: str, mkdir: bool = True):
        """
        Downloads a file if it does not exist, has zero size, or is older than the remote file.

        Args:
            url (str): The URL to download the file from.
            target_path (str): The path where the file should be saved.
            mkdir (bool): Whether to create the target directory if it does not exist. Default is False.
        """
        file_exists = os.path.exists(target_path)
        file_is_empty = file_exists and os.path.getsize(target_path) == 0
        local_mtime = os.path.getmtime(target_path) if file_exists else 0

        response = requests.head(url)
        response.raise_for_status()

        remote_mtime = response.headers.get("Last-Modified")
        if remote_mtime:
            remote_mtime = time.mktime(
                time.strptime(remote_mtime, "%a, %d %b %Y %H:%M:%S %Z")
            )
        else:
            remote_mtime = None

        needs_download = (
            not file_exists
            or file_is_empty
            or (remote_mtime and remote_mtime > local_mtime)
        )

        if needs_download:
            response = requests.get(url, stream=True)
            response.raise_for_status()

            if mkdir:
                os.makedirs(os.path.dirname(target_path), exist_ok=True)

            with open(target_path, "wb") as f:
                for chunk in response.iter_content(chunk_size=8192):
                    f.write(chunk)

            if remote_mtime:
                os.utime(target_path, (remote_mtime, remote_mtime))



"""
Created on 2025-02-24

@author: wf
"""

import os

from lodstorage.sql import SQLDB
from ngwidgets.profiler import Profiler

from genwiki.genwiki_paths import GenWikiPaths
from genwiki.multilang_querymanager import MultiLanguageQueryManager


class DjVuManager:
    """
    manager for DjVu files
    """

    def __init__(self, db_path: str = None):
        examples_path = GenWikiPaths.get_examples_path()
        yaml_path = os.path.join(examples_path, "djvu_queries.yaml")
        self.mlqm = MultiLanguageQueryManager(yaml_path=yaml_path)
        if db_path is None:
            db_path = os.path.join(examples_path, "djvu_data.db")
        self.sql_db = SQLDB(db_path, check_same_thread=False)

    def query(self, query_name: str):
        query = self.mlqm.query4Name(query_name)
        lod = self.sql_db.query(query.query)
        return lod

    def store(self, lod, entity_name: str, primary_key: str, profile: bool = True):
        """
        store my the given list of dicts
        """
        profiler = Profiler(f"caching {entity_name} to SQL", profile=profile)
        self.entity_info = self.sql_db.createTable(
            listOfRecords=lod,
            entityName=entity_name,
            primaryKey=primary_key,
            withCreate=True,
            withDrop=True,
            sampleRecordCount=20,
        )
        self.sql_db.store(
            listOfRecords=lod,
            entityInfo=self.entity_info,
            executeMany=True,
            fixNone=True,
        )
        profiler.time()



"""
Created on 2024-08-15

@author: wf
"""

import os
import unittest

from tqdm import tqdm

from genwiki.addressbook import AddressBookConverter
from genwiki.wiki import Wiki
from tests.gbasetest import GenealogyBasetest


class TestCategories(GenealogyBasetest):
    """
    test category handling

    you need to run

    wikibackup -s genealogy -q "[[Kategorie:Adressbuch_in_der_Online-Erfassung/fertig]]"

    """

    def setUp(self, debug=False, profile=True):
        GenealogyBasetest.setUp(self, debug=debug, profile=profile)
        self.wiki_id = "genealogy"
        self.get_wiki_user(wikiId=self.wiki_id, save=self.inPublicCI())
        self.target_wiki_id = "gensmw"
        self.get_wiki_user(wikiId=self.target_wiki_id, save=self.inPublicCI())
        self.target_wiki = Wiki(wiki_id=self.target_wiki_id, debug=self.debug)
        self.target_wiki.wiki_push.toWiki = self.target_wiki.wiki_push.fromWiki
        self.ask_query = "[[Kategorie:Adressbuch_in_der_Online-Erfassung/fertig]]"
        self.backup_dir = "/tmp/genwiki"
        self.wiki = Wiki(
            wiki_id=self.wiki_id, debug=self.debug, backup_dir=self.backup_dir
        )
        self.wiki.wiki_push.fromWiki.is_smw_enabled = False
        self.wiki.wiki_push.toWiki = self.target_wiki.wiki_push.fromWiki

    def test_query(self):
        """
        test SMW Ask query access
        """
        ask_query = """{{#ask: [[Concept:AddressBook]]
    |mainlabel=AddressBook
    |?AddressBook id = id
    |?AddressBook title = title
    |?AddressBook subtitle = subtitle
    |?AddressBook year = year
    |?AddressBook pageCount = pageCount
    |?AddressBook location = location
    |limit=20
    }}"""
        qlod = self.target_wiki.query_as_list_of_dicts(ask_query)
        self.assertIsInstance(qlod, list)
        self.assertTrue(len(qlod) > 0)
        debug = self.debug
        # debug=True
        for record in qlod:
            self.assertIsInstance(record, dict)
            self.assertTrue("title" in record)
            if debug:
                print(record)
        pass

    def testCategories(self):
        """
        Test categories by processing all .wiki files in the backup directory.
        """
        with_backup = self.inPublicCI() or not os.path.exists(self.backup_dir)
        # with_backup=True
        if with_backup:
            self.wiki.backup(ask_query=self.ask_query)
        page_contents = self.wiki.get_all_content()
        self.assertTrue(len(page_contents) > 10)
        if self.debug:
            print(page_contents)

    @unittest.skipIf(GenealogyBasetest.inPublicCI(), "avoid wiki writing")
    def test_convert_addressbooks(self):
        """
        test converting address books
        """
        # recreate all entries
        force = False
        # force = False

        # Set up source and target wikis
        source_wiki = self.wiki

        # Create AddressBookConverter
        ac = AddressBookConverter(force=force, debug=self.debug)

        # Get page contents
        page_contents = source_wiki.get_all_content()

        limit = 500

        # Set up tqdm progress bar
        progress_bar = tqdm(total=limit, desc="Converting AddressBooks", unit="page")

        mode = "push"  # or 'push' if you want to test pushing to the wiki

        # Convert with limit of 10 pages
        target_pages = ac.convert(
            page_contents,
            source_wiki=source_wiki,
            target_wiki=self.target_wiki,
            mode=mode,
            limit=limit,
            force=force,
            progress_bar=progress_bar,
        )

        # Close the progress bar
        progress_bar.close()

        rcount = len(target_pages)
        # Assertions
        self.assertEqual(
            progress_bar.n, rcount, f"Expected {rcount} pages to be processed"
        )

        # Check if backup files were created (if mode is 'backup')
        if mode == "backup":
            ab_count = 0
            page_contents = self.target_wiki.get_all_content()
            for _page_name, page_content in page_contents.items():
                ab_dict = ac.template_map.as_topic_dict(page_content)
                if ab_dict:
                    ab_count += 1

            self.assertTrue(
                rcount - ab_count < 3,
                f"Expected {rcount} {ac.template_map.topic_name} backup files to be created -max 3 difference",
            )

        if self.debug:
            print(f"Converted {rcount} AddressBook pages")

    def test_convert_template(self):
        """
        Test the template conversion functionality for both SiDIF and markup formats.
        """
        ac = AddressBookConverter()
        template_map = ac.template_map
        page_content = """
        {{Info Adressbuch
        | Bild = Weimar-AB-1851_Cover.jpg
        | Titel = Adreß-Buch der Residenz-Stadt Weimar
        | Untertitel = Ein Handbuch für Einheimische und Fremde
        | Autor / Hrsg. = C. G. Kästner
        | Erscheinungsjahr = 1851
        | Seitenzahl = 50
        | Enthaltene Orte = [[GOV:WEIMARJO50QX|Weimar]]
        | GOV-Quelle = source_1134489
        | Standort online = {{ThULB|http://zs.thulb.uni-jena.de/receive/jportal_jpvolume_00103933}}
        | DES = weimarTH1851
        }}
        """
        page_name = "Weimar/Adressbuch 1851"
        dict_result = template_map.as_template_dict(
            page_name=page_name, page_content=page_content
        )
        if self.debug:
            print(dict_result)

        # Test SiDIF conversion
        sidif_result = template_map.convert_template(
            page_name=page_name, page_content=page_content, output_format="sidif"
        )
        expected_sidif_lines = [
            "AddressBook1 isA AddressBook",
            '"Weimar-AB-1851_Cover.jpg" is image of AddressBook1',
            '"Adreß-Buch der Residenz-Stadt Weimar" is title of AddressBook1',
            '"Ein Handbuch für Einheimische und Fremde" is subtitle of AddressBook1',
            '"C. G. Kästner" is author of AddressBook1',
            '"1851" is year of AddressBook1',
            '"50" is pageCount of AddressBook1',
            '"WEIMARJO50QX" is location of AddressBook1',
            '"source_1134489" is govSource of AddressBook1',
            '"http://zs.thulb.uni-jena.de/receive/jportal_jpvolume_00103933" is onlineSource of AddressBook1',
            '"weimarTH1851" is des of AddressBook1',
        ]

        if self.debug:
            print("SiDIF Conversion Result:")
            print(sidif_result)
            print("\nExpected SiDIF Result:")
            print("\n".join(expected_sidif_lines))

        for line in expected_sidif_lines:
            self.assertIn(
                line, sidif_result, f"Expected line not found in SiDIF result: {line}"
            )

        # Test markup conversion
        markup_result = template_map.convert_template(
            page_name=page_name, page_content=page_content, output_format="markup"
        )
        expected_markup_lines = [
            "{{AddressBook",
            "| image = Weimar-AB-1851_Cover.jpg",
            "| title = Adreß-Buch der Residenz-Stadt Weimar",
            "| subtitle = Ein Handbuch für Einheimische und Fremde",
            "| author = C. G. Kästner",
            "| year = 1851",
            "| pageCount = 50",
            "| location = WEIMARJO50QX",
            "| govSource = source_1134489",
            "| onlineSource = http://zs.thulb.uni-jena.de/receive/jportal_jpvolume_00103933",
            "| des = weimarTH1851",
            "}}",
        ]

        if self.debug:
            print("\nMarkup Conversion Result:")
            print(markup_result)
            print("\nExpected Markup Result:")
            print("\n".join(expected_markup_lines))

        for line in expected_markup_lines:
            self.assertIn(
                line, markup_result, f"Expected line not found in markup result: {line}"
            )

        # Test that integer conversion worked
        self.assertIn('"1851" is year of AddressBook1', sidif_result)
        self.assertIn('"50" is pageCount of AddressBook1', sidif_result)
        self.assertIn("| year = 1851", markup_result)
        self.assertIn("| pageCount = 50", markup_result)

        # Test that regex extraction worked
        self.assertIn('"WEIMARJO50QX" is location of AddressBook1', sidif_result)
        self.assertIn("| location = WEIMARJO50QX", markup_result)

        if self.debug:
            print("\nAll assertions passed successfully.")



"""
Created on 2024-08-15

@author: wf
"""

import os

from lodstorage.sql import SQLDB
from ngwidgets.input_webserver import InputWebserver, InputWebSolution
from ngwidgets.login import Login
from ngwidgets.profiler import Profiler
from ngwidgets.users import Users
from ngwidgets.webserver import WebserverConfig
from ngwidgets.widgets import Link
from nicegui import Client, app, ui
from starlette.responses import FileResponse, HTMLResponse, RedirectResponse
from wd.wditem_search import WikidataItemSearch

from genwiki.convert import ParquetAdressbokToSql
from genwiki.djvu_viewer import DjVuViewer
from genwiki.djvu_catalog import DjVuCatalog
from genwiki.genwiki_paths import GenWikiPaths
from genwiki.multilang_querymanager import MultiLanguageQueryManager
from genwiki.query_view import QueryView
from genwiki.version import Version
from genwiki.wiki import Wiki
from genwiki.wikidata_view import WikidataItemView


class GenWikiWebServer(InputWebserver):
    """WebServer class that manages the server and handles GenWiki operations."""

    @classmethod
    def get_config(cls) -> WebserverConfig:
        copy_right = "(c)2024-2025 Wolfgang Fahl"
        config = WebserverConfig(
            copy_right=copy_right,
            version=Version(),
            default_port=9852,
            short_name="genwiki2024",
        )
        server_config = WebserverConfig.get(config)
        server_config.solution_class = GenWikiSolution
        return server_config

    def examples_path(self) -> str:
        path = GenWikiPaths.get_examples_path()
        return path

    def __init__(self):
        """Constructs all the necessary attributes for the WebServer object."""
        InputWebserver.__init__(self, config=GenWikiWebServer.get_config())
        users = Users(self.config.base_path)
        self.login = Login(self, users)
        self.djvu_viewer = DjVuViewer(app=app)

        address_db_path = os.path.join(self.config.storage_path, "address.db")
        if os.path.isfile(address_db_path) and os.path.getsize(address_db_path) > 0:
            self.sql_db = SQLDB(address_db_path, check_same_thread=False)
        else:
            profiler = Profiler("convert parquet files", profile=True)
            pats = ParquetAdressbokToSql(folder=self.examples_path())
            self.sql_db = SQLDB(address_db_path, check_same_thread=False)
            pats.to_db(self.sql_db)
            profiler.time()

        yaml_path = os.path.join(self.examples_path(), "queries.yaml")
        self.mlqm = MultiLanguageQueryManager(yaml_path=yaml_path)

        @ui.page("/")
        async def home(client: Client):
            return await self.page(client, GenWikiSolution.home)

        @ui.page("/wd/{qid}")
        async def wikidata_item(client: Client, qid: str):
            """
            show the wikidata item for the given Wikidata QID
            """
            await self.page(client, GenWikiSolution.wikidata_item, qid)

        @ui.page("/wd")
        async def wikidata(client: Client):
            return await self.page(client, GenWikiSolution.wikidata)

        @ui.page("/login")
        async def login(client: Client):
            return await self.page(client, GenWikiSolution.login_ui)

        @ui.page("/logout")
        async def logout(client: Client) -> RedirectResponse:
            if self.login.authenticated():
                await self.login.logout()
            return RedirectResponse("/")

        @ui.page("/djvu/catalog")
        async def djvu_catalog(client: Client):
            return await self.page(client, GenWikiSolution.djvu_catalog)  # Add route

        @app.get("/djvu/content/{file:path}")
        def get_content(file: str) -> FileResponse:
            """
            Serves content from a wrapped DjVu file.

            Args:
                file (str): The full path  <DjVu name>/<file name>.

            Returns:
                FileResponse: The requested content file (PNG, JPG, YAML, etc.).
            """
            file_response = self.djvu_viewer.get_content(file)
            return file_response

        @app.get("/djvu/{path:path}")
        def display_djvu(path: str, page: int = 1) -> HTMLResponse:
            """
            Fetches and displays a specific PNG page of a DjVu file.
            """
            html_response = self.djvu_viewer.get_page(path, page)
            return html_response

    def configure_run(self):
        super().configure_run()
        self.wiki_id = "gensmw"
        self.wiki = Wiki(wiki_id=self.wiki_id, debug=self.args.debug)


class GenWikiSolution(InputWebSolution):
    """
    the GenWiki solution
    """

    def __init__(self, webserver: GenWikiWebServer, client: Client):
        """
        Initialize the solution

        Args:
            webserver (GenWikiWebServer): The webserver instance associated with this solution.
            client (Client): The client instance this context is associated with.
        """
        super().__init__(webserver, client)
        self.mlqm = webserver.mlqm
        self.wiki = webserver.wiki
        self.sql_db = self.webserver.sql_db

    def authenticated(self) -> bool:
        """
        Check if the user is authenticated.
        Returns:
            True if the user is authenticated, False otherwise.
        """
        return self.webserver.login.authenticated()

    def setup_menu(self, detailed: bool = True):
        """
        setup the menu
        """
        super().setup_menu(detailed=detailed)
        with self.header:
            self.link_button("DjVu Catalog", "/djvu/catalog", "library_books")  # Add menu entry
            if self.authenticated():
                self.link_button("logout", "/logout", "logout", new_tab=False)
            else:
                self.link_button("login", "/login", "login", new_tab=False)

    async def login_ui(self):
        """
        login ui
        """
        await self.webserver.login.login(self)

    async def home(self):
        """Provide the main content page"""

        def setup_home():
            """ """
            self.query_view = QueryView(
                self, mlqm=self.mlqm, sql_db=self.sql_db, wiki=self.wiki
            )
            self.query_view.setup_ui()

        await self.setup_content_div(setup_home)

    async def wikidata_item(self, qid: str):
        """
        show a wikidata item with the given Wikidata id

        Args:
            qid(str): the Wikidata id of the item to show
        """

        def show():
            self.wdv = WikidataItemView(self, mlqm=self.mlqm, qid=qid)
            self.wdv.setup_ui()

        await self.setup_content_div(show)

    async def wikidata(self):
        """
        provide the location page
        """

        def record_filter(qid: str, record: dict):
            if "label" and "desc" in record:
                text = f"""{record["label"]}({qid})☞{record["desc"]}"""
                wd_link = Link.create(f"/wd/{qid}", text)
                # getting the link to be at second position
                # is a bit tricky
                temp_items = list(record.items())
                # Add the new item in the second position
                temp_items.insert(1, ("wikidata item", wd_link))

                # Clear the original dictionary and update it with the new order of items
                record.clear()
                record.update(temp_items)

        def show():
            self.wd_item_search = WikidataItemSearch(
                self, record_filter=record_filter, lang="de"
            )

        await self.setup_content_div(show)

    async def djvu_catalog(self):
        """Show the DjVu Catalog page"""

        def show():
            self.djvu_catalog_view = DjVuCatalog(self)
            self.djvu_catalog_view.setup_ui()

        await self.setup_content_div(show)




"""
Created on 14.09.2024

@author: wf
"""

import re

from nicegui import ui

from genwiki.multilang_querymanager import MultiLanguageQueryManager
from genwiki.wikidata import Wikidata


class WikidataItemView:
    """
    show a wikidata item
    """

    def __init__(self, solution, mlqm: MultiLanguageQueryManager, qid: str):
        self.solution = solution
        self.mlqm = mlqm

        self.qid = Wikidata.unprefix(qid)
        self.item_query = self.mlqm.query4Name("WikidataItemNameAndCoordinates")

    def convert_point_to_latlon(self, point_str: str):
        # Define the pattern to match the "Point(x y)" format
        pattern = r"Point\(([-\d.]+)\s+([-\d.]+)\)"

        # Define the replacement pattern
        replacement = r"\1°, \2°"

        # Use re.sub to replace the match
        return re.sub(pattern, replacement, point_str)

    def setup_ui(self):
        """
        setup the user interface
        """
        ui.label(self.qid)
        query = self.item_query
        query.params.params_dict["item"] = self.qid
        sparql = Wikidata.get_sparql()
        qlod = sparql.queryAsListOfDicts(
            query.query, param_dict=query.params.params_dict
        )
        if len(qlod) == 1:
            record = qlod[0]
            wikidataid = record["item"]
            wikidataid = Wikidata.unprefix(wikidataid)
            if "coordinates" in record:
                point_str = record["coordinates"]
                latlon = self.convert_point_to_latlon(point_str)
                coord = f"\n|coordinates={latlon}"
            else:
                coord = ""
                pass
            # convert to -32.715°, -77.03201°
            wiki_markup = f"""{{{{Location
|name={record["itemLabel"]}
|wikidataid={wikidataid}{coord}
}}}}"""
            html_markup = f"<pre>{wiki_markup}</pre>"
            ui.html(html_markup)
        else:
            ui.notify(f"Could not retrieve details for {self.qid}")
        pass



"""
Created on 26.08.2024

@author: wf
"""
import os
from genwiki.query_view import QueryView  # Assuming QueryView is in `genwiki.query_view`
from genwiki.djvu_manager import DjVuManager
from genwiki.multilang_querymanager import MultiLanguageQueryManager
from ngwidgets.widgets import Link

class DjVuCatalog(QueryView):
    """
    UI for browsing and querying the DjVu document catalog.
    """

    def __init__(self, solution):
        self.solution=solution
        self.webserver=self.solution.webserver
        storage_path=solution.webserver.config.storage_path
        db_path=os.path.join(storage_path,"genwiki_djvu.db")
        yaml_path = os.path.join(self.webserver.examples_path(), "djvu_queries.yaml")
        self.mlqm = MultiLanguageQueryManager(yaml_path=yaml_path)
        try:
            self.dvm = DjVuManager(db_path=db_path)
            super().__init__(solution=solution,
                mlqm=self.mlqm,
                sql_db=self.dvm.sql_db,
                wiki=self.webserver.wiki)
            self.query_name = "all_djvu"
        except Exception as ex:
            self.solution.handle_exception(ex)

    def get_view_record(self, record: dict, index: int) -> dict:
        """
        get the view record for the given record

        Convert path entries to two urls:
            1. GenWiki URL: Links to the file on the Genealogy Wiki.
            2. Local URL: Links to the local DjVu viewer.

        Args:
            record (dict): The source record to convert
            index (int): The row index

        Returns:
            dict: The view record with formatted links and properties
        """
        view_record = {"#": index}  # Number first
        record_copy = record.copy()
        for key, value in record_copy.items():
            if isinstance(value, str) and value.startswith("/images/"):
                filename = value.split("/")[-1]
                wiki_url = f"https://wiki.genealogy.net/index.php?title=Datei%3A{filename}"
                local_url = f"/djvu/{filename}"
                view_record["wiki"]=Link.create(url=wiki_url,text=filename)
                view_record["view"]=Link.create(url=local_url,text=filename)
            else:
                view_record[key] = value
            pass
        return view_record

    def get_view_lod(self, lod: list) -> list:
        """
        Convert records to view format with row numbers and links
        """
        view_lod = []
        for i, record in enumerate(lod):
            index = i + 1
            view_record = self.get_view_record(record, index)
            view_lod.append(view_record)
        return view_lod

    def get_query_lod(self):
        """
        Fetches DjVu catalog data based on the selected query.
        """
        self.lod=self.dvm.query(self.query_name)
        self.view_lod=self.get_view_lod(self.lod)
        return self.view_lod

    def setup_ui(self):
        """
        Sets up the UI components for the DjVu catalog.
        """
        super().setup_ui(query_name=self.query_name)
        self.query_select.options = ["all_djvu", "bundled_djvu", "unbundled_djvu"]



"""
Created on 15.09.2024

@author: wf
"""

import logging
import os
from collections import Counter
from typing import Any, Dict, List

import geocoder
from ez_wikidata.wdsearch import WikidataSearch
from geopy.distance import geodesic

from genwiki.genwiki_paths import GenWikiPaths
from genwiki.gov_api import GOV_API
from genwiki.multilang_querymanager import MultiLanguageQueryManager
from genwiki.nominatim import NominatimWrapper
from genwiki.wikidata import Wikidata


class Locator:
    """
    find locations
    """

    def __init__(self, debug: bool = False):
        self.nominatim = NominatimWrapper(user_agent="GenWiki2024LocationTest")
        self.wds = WikidataSearch()
        self.gov_api = GOV_API()
        self.sparql = Wikidata.get_sparql()
        yaml_path = os.path.join(GenWikiPaths.get_examples_path(), "queries.yaml")
        self.mlqm = MultiLanguageQueryManager(yaml_path=yaml_path)
        self.lookup_query = self.mlqm.query4Name("WikidataLookup")
        self.limit = 11
        self.lang_map = {"deu": "de", "pol": "pl"}
        self.debug = debug

    def multi_item_query(
        self, query_name, items: List[str], lang: str = "en"
    ) -> List[Dict[str, Any]]:
        """
        perform a multi item query for the given query_name with
        """
        query = self.mlqm.query4Name(query_name)
        items_str = " ".join([f"wd:{item}" for item in items if item])
        param_dict = {"items": items_str, "lang": lang}
        sparql_query = query.params.apply_parameters_with_check(param_dict)
        qlod = self.sparql.queryAsListOfDicts(
            queryString=sparql_query, param_dict=param_dict
        )
        return qlod

    def get_coordinates(self, items: List[str]) -> Dict[str, tuple]:
        """
        Get coordinates for multiple Wikidata items
        """
        qlod = self.multi_item_query(query_name="WikidataItemsCoordinates", items=items)

        coordinates = {}
        for record in qlod:
            item = Wikidata.unprefix(record["item"])
            coord_str = record["coordinates"]
            lon, lat = map(float, coord_str.strip("Point()").split())
            coordinates[item] = (lat, lon)
        return coordinates

    def lookup_item(self, item: str, lang: str = "de"):
        param_dict = {"item": item, "lang": lang}
        query = self.lookup_query
        sparql_query = query.params.apply_parameters_with_check(param_dict)
        qlod = self.sparql.queryAsListOfDicts(
            queryString=sparql_query, param_dict=param_dict
        )
        return qlod

    def to_path(self, qlod) -> str:
        path = None
        for record in qlod:
            level = record["level"]
            if level == "4":
                iso_code = record["iso_code"]
                label = record["itemLabel"]
                path = f"""{iso_code.replace("-","/")}/{label}"""
        return path

    def lookup_path_for_item(self, item: str, lang: str = "de") -> str:
        """
        lookup the path for the given item wikidata Q-Identifier
        """
        qlod = self.lookup_item(item, lang)
        path = self.to_path(qlod)
        return path

    def lookup_wikidata_id_by_geoid(
        self, geoid_kind: str, geo_id: str, lang: str = "en"
    ) -> str:
        result = None
        if geoid_kind == "NUTS2003" or geoid_kind == "NUTS1999":
            query_name = "WikidataLookupByNutsCode"
            param_name = "nuts_code"
        elif geoid_kind == "geonames":
            query_name = "WikidataLookupByGeoNamesID"
            param_name = "geonames_id"
        else:
            raise ValueError(f"invalid geo_id_kind {geoid_kind}")
        query = self.mlqm.query4Name(query_name)
        param_dict = {param_name: geo_id, "lang": lang}
        sparql_query = query.params.apply_parameters_with_check(param_dict)
        qlod = self.sparql.queryAsListOfDicts(
            queryString=sparql_query, param_dict=param_dict
        )
        if len(qlod) == 1:
            record = qlod[0]
            item = record["item"]
            wikidataid = Wikidata.unprefix(item)
            result = wikidataid
        elif len(qlod) > 1:
            raise ValueError(f"wikidata has multiple entries for {param_name}:{geo_id}")
        return result

    def locate_by_name(self, name: str, language: str = "de"):
        self.wds.language = language
        sr = self.wds.searchOptions(name, limit=self.limit)
        for j, q_record in enumerate(sr):
            qid, qlabel, desc = q_record
            if self.debug:
                print(f"{j:3}:{qlabel}({qid}):{desc}")
            if name in desc or name in qlabel:
                return qid
        return None

    def validate(self, gov_obj, items: Dict[str, str], max_distance_km: float = 3.0):
        """
        validate the list of wikidata items against the given gov_obj
        """
        if not gov_obj:
            return
        gov_position = gov_obj.get("position", {})
        gov_lat, gov_lon = gov_position.get("lat"), gov_position.get("lon")
        items_to_remove = []
        wikidata_coords = {}
        if not gov_lat or not gov_lon:
            msg = "Gov object does not have valid coordinates"
            logging.warn(msg)
            gov_coords = None
        else:
            gov_coords = (gov_lat, gov_lon)
            wikidata_items = [item for item in items.values() if item]
            wikidata_coords = self.get_coordinates(wikidata_items)

        for key, item in items.items():
            ok = item is not None
            if item in wikidata_coords:
                item_coords = wikidata_coords[item]
                distance = geodesic(gov_coords, item_coords).kilometers
                ok = distance <= max_distance_km
                if self.debug:
                    check_mark = "✓" if ok else "✗"
                    print(f"{key}: {distance:.1f} km {check_mark}")
            if not ok:
                items_to_remove.append(key)

        for key in items_to_remove:
            items.pop(key)

    def sort_items(self, items):
        # Count occurrences of each value
        value_counts = Counter(items.values())

        # Sort items by value, then by frequency of value
        sorted_items = sorted(items.items(), key=lambda x: (-value_counts[x[1]], x[1]))

        # Clear and update the original dictionary
        items.clear()
        items.update(sorted_items)

        return items

    def locate(self, gov_id: str) -> Dict[str, str]:
        """
        locate the locaction described by the given gov_id
        """
        items = {}
        gov_obj = None
        try:
            gov_obj = self.gov_api.get_raw_gov_object(gov_id)
            if "externalReference" in gov_obj:
                for i, ref in enumerate(gov_obj["externalReference"]):
                    val = ref["value"]
                    if self.debug:
                        print(f"{i}:{val}")
                    for geoid_kind in ["geonames", "NUTS2003", "NUTS1999"]:
                        if val.startswith(geoid_kind):
                            geo_id = val.split(":")[1]
                            item = self.lookup_wikidata_id_by_geoid(geoid_kind, geo_id)
                            if item:
                                items[val] = item
                    for name_record in gov_obj["name"]:
                        lang = name_record["lang"]
                        name = name_record["value"]
                        if lang in self.lang_map:
                            language = self.lang_map[lang]
                        else:
                            language = "en"
                        item = self.locate_by_name(name, language)
                        items[f"gov-{name}@{language}"] = item

        except Exception as ex:
            if "404" in str(ex) or "501" in str(ex):
                item = self.nominatim.lookup_wikidata_id(gov_id)
                if item:
                    items["nominatim"] = item
                pass
            else:
                raise ex
        self.validate(gov_obj, items)
        self.sort_items(items)
        return items



"""
Created on 06.05.2024

@author: wf
"""

from lodstorage.params import Params
from ngwidgets.dict_edit import DictEdit, FieldUiDef, FormUiDef


class ParamsView:
    """
    a view for Query Parameters
    """

    def __init__(
        self, solution, params: Params, title: str = "Params", icon: str = "tune"
    ):
        """
        construct me with the given solution and params
        """
        self.solution = solution
        self.params = params
        self.dict_edit = None
        self.title = title
        self.icon = icon

    def open(self):
        """
        show the details of the dict edit
        """
        self.dict_edit.open()

    def close(self):
        """
        hide the details of the dict edit
        """
        self.dict_edit.close()

    def delete(self):
        self.dict_edit.card.delete()
        self.dict_edit = None

    def setup(self):
        """
        Update the params and refresh the dict_edit
        """
        if self.dict_edit:
            self.dict_edit.card.delete()
        self.dict_edit = self.get_dict_edit()

    def get_dict_edit(self) -> DictEdit:
        """
        Return a DictEdit instance for editing parameters.
        """
        # Define a custom form definition for the title "Params"
        form_ui_def = FormUiDef(
            title=self.title,
            icon=self.icon,
            ui_fields={
                key: FieldUiDef.from_key_value(key, value)
                for key, value in self.params.params_dict.items()
            },
        )
        self.dict_edit = DictEdit(
            data_to_edit=self.params.params_dict, form_ui_def=form_ui_def
        )
        return self.dict_edit



"""
Created on 19.08.2024

@author: wf
"""

import json
import logging
import os
from typing import Any, Dict, List

from genwiki.genwiki_paths import GenWikiPaths
from genwiki.locator import Locator
from genwiki.template import TemplateMap, TemplateParam
from genwiki.wikidata import Wikidata


class AddressBookConverter:
    """
    convert Adressbook pages
    """

    def __init__(self, force: bool = False, debug: bool = False):
        self.force = force
        self.debug = debug
        self.template_map = TemplateMap(
            template_name="Info Adressbuch",
            topic_name="AddressBook",
            param_mapping={
                "Bild": TemplateParam(new_name="image"),
                "Titel": TemplateParam(new_name="title"),
                "Untertitel": TemplateParam(new_name="subtitle"),
                "Autor / Hrsg.": TemplateParam(new_name="author"),
                "Erscheinungsjahr": TemplateParam(new_name="year", to_int=True),
                "Seitenzahl": TemplateParam(new_name="pageCount", to_int=True),
                "Enthaltene Orte": TemplateParam(
                    new_name="location", regex=r"\[\[GOV:(.*?)\|"
                ),
                "Verlag": TemplateParam(new_name="publisher"),
                "GOV-Quelle": TemplateParam(new_name="govSource"),
                "Standort online": TemplateParam(
                    new_name="onlineSource", regex=r"\{\{ThULB\|(.*?)\}\}"
                ),
                "DES": TemplateParam(new_name="des"),
            },
        )
        self.year_mapping = {"weimarTH1851.parquet": 1851, "weimarTH1853.parquet": 1853}
        self.locator = Locator(debug=self.debug)
        self.target_wiki = None
        self.region_lookup = self.get_region_lookup()

    def get_region_lookup(self):
        region_lookup_path = os.path.join(
            GenWikiPaths.get_examples_path(), "region_lookup.json"
        )

        if os.path.exists(region_lookup_path):
            with open(region_lookup_path, "r") as f:
                return json.load(f)
        else:
            raise Exception(f"missing region lookup json file {region_lookup_path}")

    def create_location_page(
        self,
        page_title: str,
        item: str,
        name: str,
        partOf: str,
        level: str = "5",
        force: bool = False,
    ):
        """
        create a wiki page for the given location
        """
        if not self.target_wiki:
            raise ValueError("target wiki is not set")
        wiki = self.target_wiki.wiki_push.toWiki
        page = wiki.get_page(page_title)
        create = force or not page.exists
        kind_map = {"3": "Country", "4": "Region", "5": "City"}
        location_kind = kind_map[level]
        if create:
            coords = self.locator.get_coordinates([item])
            if not item in coords:
                logging.warn(f"no coordinates for {item}")
                coord_markup = ""
            else:
                lat, lon = coords[item]
                coord_markup = f"{lat},{lon}"
            markup = f"""{{{{Location
|path={page_title}
|name={name}
|coordinates={coord_markup}
|wikidataid={item}
|locationKind={location_kind}
|level={level}
|partOf={partOf}
}}}}
{{{{Location
|viewmode=masterdetail
}}}}
"""
            page.edit(markup, "created by genwiki2024 addressbook converter")

    def check_location(
        self,
        page_title: str,
        item,
        lookup_qlod: List[Dict[str, Any]],
        force: bool = False,
    ):
        """
        check the location page
        """
        if not page_title:
            msg = f"missing page title for item {item}"
            logging.error(msg)
            return
        parts = page_title.rsplit("/", 1)  # Split from the right, limit to one split
        partOf = parts[0]  # PL/22
        name = parts[1]  # Miastko
        # PL/22/Miastko
        self.create_location_page(
            page_title=page_title, item=item, name=name, partOf=partOf, force=force
        )
        for record in lookup_qlod:
            level = record["level"]
            iso_code = record["iso_code"]
            item = record["intermediateAdmin"]
            item = Wikidata.unprefix(item)
            name = record["intermediateAdminLabel"]
            page_title = f"""{iso_code.replace("-","/")}"""
            parts = page_title.rsplit(
                "/", 1
            )  # Split from the right, limit to one split
            partOf = parts[0]  # PL/22
            # country level
            if level == "3":
                if item in self.region_lookup:
                    lookup = self.region_lookup[item]
                    partOf = lookup["itemLabel"]
                else:
                    partOf = ""
                pass
            self.create_location_page(
                page_title=page_title,
                item=item,
                name=name,
                partOf=partOf,
                level=level,
                force=force,
            )

    def record_convert(self, page_name: str, page_content: str, record: Dict):
        """
        addressbook record conversion callback
        """
        gov_id = record["location"]
        page_title = page_name.replace(" ", "_")
        record["id"] = page_title
        record["genwikiurl"] = f"https://wiki.genealogy.net/{page_title}"
        if "image" in record and record["image"]:
            record["image"] = f"""File:{record["image"]}"""
        items_dict = self.locator.locate(gov_id)
        if len(items_dict) > 0:
            for key, item in items_dict.items():
                if not item:
                    continue
                lookup_qlod = self.locator.lookup_item(item)
                location_title = self.locator.to_path(lookup_qlod)
                if location_title:
                    record["at"] = location_title
                    self.check_location(
                        location_title, item, lookup_qlod, force=self.force
                    )
                    break

    def convert(
        self,
        page_contents,
        source_wiki=None,
        target_wiki=None,
        mode="backup",
        limit: int = None,
        dry_run: bool = False,
        force: bool = False,
        progress_bar=None,
    ):
        """
        Convert the pages of the given page_contents dict.

        Args:
            page_contents (dict): Dictionary of page names and their content.
            target_wiki (Wiki): Target wiki object for pushing content (if mode is 'push').
            mode (str): 'push' to push to target wiki, 'backup' to store as backup files.
            limit (int): Limit the number of pages to process.
            force (bool): if not True only do a dry run
            progress_bar (Progressbar): Progress bar object to update during conversion.
        """
        self.target_wiki = target_wiki
        if not dry_run:
            target_wiki.wiki_push.toWiki.login()
        result = {}
        total = len(page_contents) if limit is None else min(limit, len(page_contents))
        if progress_bar:
            progress_bar.total = total
            progress_bar.set_description(f"Converting {total} pages")
        wiki = target_wiki.wiki_push.toWiki
        for i, (page_name, page_content) in enumerate(page_contents.items()):
            if limit and i >= limit:
                break
            # avoid effort for existing pages if not in force mode
            if mode == "push" and target_wiki:
                page = wiki.get_page(page_name)
                if page.exists and not force:
                    logging.info(f"page {page_name} exists")
                    continue

            ab_dict = self.template_map.as_template_dict(
                page_name, page_content, callback=self.record_convert
            )
            markup = self.template_map.dict_to_markup(ab_dict)
            result[page_name] = markup
            if mode == "push" and target_wiki:
                edit_status = target_wiki.wiki_push.edit_page_content(
                    page_title=page_name,
                    new_text=markup,
                    summary="Converted AddressBook template",
                    force=not dry_run,
                )
                logging.log(logging.INFO, edit_status)
                pass
                source_page = source_wiki.wiki_push.fromWiki.getPage(page_name)
                image = None
                for page_image in source_page.images():
                    base_name = page_image.base_name.replace("Datei:", "File:")
                    if "image" in ab_dict and base_name == ab_dict["image"]:
                        image = page_image
                    pass
                if image:
                    source_wiki.wiki_push.pushImages([image], ignore=True)
            elif mode == "backup":
                backup_path = os.path.join(
                    target_wiki.wiki_backup_dir, f"{page_name}.wiki"
                )
                os.makedirs(os.path.dirname(backup_path), exist_ok=True)
                with open(backup_path, "w", encoding="utf-8") as f:
                    f.write(markup)

            if progress_bar:
                progress_bar.update(1)

            if self.debug:
                print(f"Processed page: {page_name}")

        if progress_bar:
            progress_bar.set_description("Conversion complete")
        return result



"""
Created on 19.08.2024

@author: wf
"""

import re
from dataclasses import field
from typing import Any, Callable, Dict, Optional

import mwparserfromhell
from ngwidgets.yamlable import lod_storable


@lod_storable
class TemplateParam:
    new_name: str
    regex: Optional[str] = None
    to_int: bool = False


@lod_storable
class TemplateMap:
    template_name: str
    topic_name: str
    param_mapping: Dict[str, TemplateParam] = field(default_factory=dict)

    def as_template_dict(
        self, page_name: str, page_content: str, callback: Callable = None
    ) -> Dict[str, Any]:
        wikicode = mwparserfromhell.parse(page_content)
        templates = wikicode.filter_templates(matches=self.template_name)

        if not templates:
            return {}

        template = templates[0]
        result = {}

        for old_field, param in self.param_mapping.items():
            if template.has(old_field):
                value = str(template.get(old_field).value).strip()

                if param.regex:
                    match = re.search(param.regex, value)
                    value = match.group(1) if match else value

                if param.to_int:
                    try:
                        value = int(value)
                    except ValueError:
                        pass  # Keep the original value if it's not a valid integer

                result[param.new_name] = value
        if callback:
            callback(page_name, page_content, result)
        return result

    def as_topic_dict(self, page_content: str) -> Dict[str, Any]:
        """
        Extract template parameters based on the topic_name without conversions.

        Args:
            page_content (str): The content of the wiki page containing the template.

        Returns:
            Dict[str, Any]: Extracted parameters as a dictionary.
        """
        wikicode = mwparserfromhell.parse(page_content)
        templates = wikicode.filter_templates(matches=self.topic_name)

        if not templates:
            return {}

        template = templates[0]
        result = {}

        for param in template.params:
            name = str(param.name).strip()
            value = str(param.value).strip()
            result[name] = value

        return result

    def dict_to_sidif(self, data: Dict[str, Any]) -> str:
        sidif_lines = [f"{self.topic_name}1 isA {self.topic_name}"]
        for key, value in data.items():
            sidif_lines.append(f'"{value}" is {key} of {self.topic_name}1')
        return "\n".join(sidif_lines)

    def convert_template(
        self, page_name: str, page_content: str, output_format: str = "sidif"
    ) -> str:
        data = self.as_template_dict(page_name=page_name, page_content=page_content)
        if not data:
            return f"No '{self.template_name}' template found in the content."

        if output_format == "sidif":
            return self.dict_to_sidif(data)
        elif output_format == "markup":
            return self.dict_to_markup(data)
        else:
            raise ValueError(f"Unsupported output format: {output_format}")

    # Add dict_to_markup method if not already present
    def dict_to_markup(self, data: Dict[str, Any]) -> str:
        markup_lines = [f"{{{{{self.topic_name}"]
        for key, value in data.items():
            markup_lines.append(f"| {key} = {value}")
        markup_lines.append("}}")
        return "\n".join(markup_lines)



"""
Created on 2025-02-25

@author: wf
"""

import io
import mimetypes
import os
import tarfile
from pathlib import Path

from fastapi import FastAPI, HTTPException
from fastapi.responses import HTMLResponse, Response
from fastapi.staticfiles import StaticFiles

from genwiki.djvu_core import DjVuFile


class DjVuViewer:
    """
    Handles loading and retrieving DjVu page metadata from YAML files and sets up static file serving.
    """

    _static_mounted = False  # Ensures mount is only done once

    def __init__(self, app: FastAPI, base_path: str = None):
        if base_path is None:
            base_path = os.getenv(
                "GENWIKI_PATH", "/Users/wf/hd/wf-fur.bitplan.com/genwiki"
            )
        self.image_path = os.path.join(base_path, "djvu_images")

        if not DjVuViewer._static_mounted:
            app.mount(
                "/static/djvu",
                StaticFiles(directory=self.image_path),
                name="djvu_images",
            )
            DjVuViewer._static_mounted = True

    def read_from_tar(self, tarball_path: Path, filename: str) -> bytes:
        """
        Reads a file directly from a tarball.

        Args:
            tarball_path (Path): Path to the tar archive.
            filename (str): Name of the file inside the archive.

        Returns:
            bytes: The file contents.
        """
        with tarfile.open(tarball_path, "r") as tar:
            try:
                member = tar.getmember(filename)
                with tar.extractfile(member) as file:
                    return file.read()
            except KeyError:
                raise HTTPException(
                    status_code=404, detail=f"File {filename} not found in tarball"
                )

    def get_content(self, file: str) -> Response:
        """
        Retrieves a content file (PNG, JPG, YAML, etc.) from the tarball and serves it as a response.

        Args:
            file (str): The full path in the format <DjVu name>/<file name>.

        Returns:
            Response: The requested content file with the correct media type.
        """
        try:
            djvu_name, filename = file.split("/", 1)
        except ValueError:
            raise HTTPException(
                status_code=400,
                detail="Invalid file path format. Expected <DjVu name>/<file name>.",
            )

        tarball_path = Path(self.image_path) / f"{djvu_name}.tar"

        if not tarball_path.exists():
            raise HTTPException(status_code=404, detail="Tarball not found")

        file_content = self.read_from_tar(tarball_path, filename)
        file_stream = io.BytesIO(file_content)
        # Detect MIME type based on file extension
        media_type, _ = mimetypes.guess_type(filename)
        if media_type is None:
            media_type = "application/octet-stream"  # Default for unknown types

        content_response = Response(content=file_content, media_type=media_type)

        return content_response

    def get_page(self, path: str, page_index: int) -> HTMLResponse:
        """
        Fetches and renders an HTML page displaying the PNG image of the given DjVu file page from a tarball.
        """
        tarball_file = Path(self.image_path) / f"{Path(path).stem}.tar"
        yaml_file = f"{Path(path).stem}.yaml"

        if not tarball_file.exists():
            raise HTTPException(status_code=404, detail="Tarball not found")

        try:
            yaml_data = self.read_from_tar(tarball_file, yaml_file).decode("utf-8")
            djvu_file = DjVuFile.from_yaml(yaml_data)
        except HTTPException:
            raise HTTPException(
                status_code=404, detail="YAML metadata not found in tarball"
            )
        except Exception:
            raise HTTPException(
                status_code=500, detail="Error reading YAML from tarball"
            )

        page_count = len(djvu_file.pages)
        if page_index < 1 or page_index > page_count:
            raise HTTPException(status_code=404, detail=f"Page {page_index} not found")

        djvu_page = djvu_file.pages[page_index - 1]
        image_filename = djvu_page.png_file
        image_url = f"/djvu/content/{Path(path).stem}/{image_filename}"

        return HTMLResponse(
            content=self.get_markup(path, page_index, len(djvu_file.pages), image_url)
        )

    def get_markup(
        self, path: str, page_index: int, total_pages: int, image_url: str
    ) -> str:
        """
        Returns the HTML markup for displaying the DjVu page with navigation.

        Args:
            path (str): DjVu file path.
            page_index (int): Current page index.
            total_pages (int): Total number of pages in the DjVu document.
            image_url (str): URL to the PNG file.

        Returns:
            str: HTML markup.
        """
        first_page = 1  # Fix: Pages start from 1
        last_page = total_pages  # Fix: Last page is total_pages, not total_pages - 1
        prev_page = max(first_page, page_index - 1)
        next_page = min(last_page, page_index + 1)
        fast_backward = max(first_page, page_index - 10)
        fast_forward = min(last_page, page_index + 10)

        markup = f"""
        <!DOCTYPE html>
        <html lang="en">
        <head>
            <meta charset="UTF-8">
            <meta name="viewport" content="width=device-width, initial-scale=1.0">
            <title>DjVu Viewer</title>
            <style>
                body {{ font-family: Arial, sans-serif; text-align: center; }}
                img {{ max-width: 100%; height: auto; }}
                .nav {{ margin-top: 20px; }}
                .nav a {{ margin: 0 10px; text-decoration: none; font-weight: bold; font-size: 24px; }}
            </style>
        </head>
        <body>
            <div class="nav">
                <a href="/djvu/{path}?page={first_page}" title="First Page (1/{total_pages})">⏮</a>
                <a href="/djvu/{path}?page={fast_backward}" title="Fast Backward (Jump -10 Pages)">⏪</a>
                <a href="/djvu/{path}?page={prev_page}" title="Previous Page">⏴</a>
                <span>{page_index} / {total_pages}</span>
                <a href="/djvu/{path}?page={next_page}" title="Next Page">⏵</a>
                <a href="/djvu/{path}?page={fast_forward}" title="Fast Forward (Jump +10 Pages)">⏩</a>
                <a href="/djvu/{path}?page={last_page}" title="Last Page ({total_pages}/{total_pages})">⏭</a>
            </div>
            <img src="{image_url}" alt="DjVu Page {page_index}">
        </body>
        </html>
        """
        return markup



"""
Created on 2025-02-24

@author: wf
"""

import argparse
import json
import os

import djvu.decode
from ngwidgets.basetest import Basetest

from genwiki.djvu_cmd import DjVuCmd
from genwiki.djvu_manager import DjVuManager
from genwiki.djvu_processor import DjVuProcessor
from genwiki.download import Download


class TestDjVu(Basetest):
    """
    Test djvu handling
    """

    def setUp(self, debug=True, profile=True):

        Basetest.setUp(self, debug=debug, profile=profile)
        self.basepath = DjVuCmd.default_base_path
        self.output_dir = os.path.join(self.basepath, "djvu_images")
        self.baseurl = "https://wiki.genealogy.net/"
        self.limit = 10000000
        # self.limit=50
        self.local = True
        if not os.path.isdir(self.basepath):
            self.local = False
            self.basepath = "/tmp/genwiki/images"
            self.output_dir = "/tmp/genwiki/djvu_images"
            self.limit = 50
        os.makedirs(self.output_dir, exist_ok=True)

    def get_djvu(self, relurl):
        """
        get the djvu file for the relative url
        """
        djvu_path = self.basepath + relurl
        url = self.baseurl + relurl
        if not self.local:
            try:
                Download.download(url, djvu_path)
            except Exception as _ex:
                print(f"invalid {djvu_path}")
                return None
        self.assertTrue(os.path.isfile(djvu_path), djvu_path)
        return djvu_path

    def test_djvu_processor(self):
        """
        test the DjVu processor
        """
        for relurl,elen in [
            ("/images/2/2f/Sorau-AB-1913.djvu",255),
            ("/images/9/96/Elberfeld-AB-1896-97-Stadtplan.djvu",1),
            ("/images/9/96/vz1890-neuenhausen-zb04.djvu",3),
            ("/images/0/08/Deutsches-Kirchliches-AB-1927.djvu",1188)
        ]:
            djvu_path = self.get_djvu(relurl)
            url = djvu.decode.FileURI(djvu_path)
            # url=f"{baseurl}/{relurl}"
            dproc = DjVuProcessor()
            if self.debug:
                print(f"processing {url}")
            document = dproc.context.new_document(url)
            document.decoding_job.wait()
            if self.debug:
                page_count=len(document.files)
                print(f"{page_count} pages")
            self.assertEqual(elen,page_count)
        pass

    def test_all_djvu(self):
        """
        test all djvu pages
        """
        if self.inPublicCI():
            return
        args = argparse.Namespace(
            command="catalog",
            db_path="/tmp/test_genwiki_djvu.db",
            base_path=DjVuCmd.default_base_path,
            limit=10000000,
            sort="asc",
            force=False,
            output_path=self.output_dir,
            parallel=False,
            debug=True,
            verbose=False,
            serial=False
        )
        djvu_cmd = DjVuCmd(args=args)
        djvu_cmd.handle_args()
        expected_errors = 0 if self.local else 2
        self.assertTrue(djvu_cmd.errors <= expected_errors)

    def test_convert(self):
        """
        test the conversion
        """
        args = argparse.Namespace(
            command="convert",
            db_path="/tmp/test_genwiki_djvu.db",
            base_path=self.basepath,
            limit=50,
            force=True,
            sort="asc",
            output_path=self.output_dir,
            parallel=True,
            url="/images/2/2f/Sorau-AB-1913.djvu",
            #url="/images/9/96/vz1890-neuenhausen-zb04.djvu",
            debug=True,
            serial=False,
            verbose=True,
        )
        djvu_cmd = DjVuCmd(args=args)
        djvu_cmd.handle_args()

    def test_issue49(self):
        """
        Test loading DjVu file with python-djvu and storing relevant metadata.
        """
        output_dir = "/tmp/djvu_pngs"
        os.makedirs(output_dir, exist_ok=True)
        for url, page_count in [
            ("/images/9/96/vz1890-neuenhausen-zb04.djvu", 3),
            ("./images/f/fc/Siegkreis-AB-1905-06_Honnef.djvu", 35),
            # ("./images/9/96/Elberfeld-AB-1896-97-Stadtplan.djvu", 1),
            # ("./images/0/08/Deutsches-Kirchliches-AB-1927.djvu", 1188),
        ]:
            with self.subTest(url=url, expected_pages=page_count):
                if not self.local and page_count > 1:
                    return
                relurl = url.lstrip(".")
                djvu_path = self.get_djvu(relurl)
                dproc = DjVuProcessor()
                if self.debug:
                    print(f"processing {relurl}")
                # for document, page in dproc.yield_pages(djvu_path):
                #    pass
                count = 0
                for _image_job in dproc.process_parallel(
                    djvu_path, relurl=relurl, save_png=True, output_path=output_dir
                ):
                    count += 1

                if self.debug:
                    print(f"Processed {count} pages")

    def testDjVuManager(self):
        """
        test the DjVu Manager
        """
        dvm = DjVuManager()
        lod = dvm.query("total")
        if self.debug:
            print(json.dumps(lod, indent=2))
        self.assertEqual(lod, [{"files": 4288, "pages": 1028225}])



"""
Created on 26.08.2024

@author: wf
"""

import os

from lodstorage.sql import SQLDB

from genwiki.parquet import Parquet


class ParquetAdressbokToSql:
    """
    convert parquet Addressbook files to Sql
    """

    def __init__(
        self,
        folder: str,
        column_mapping: dict = None,
        table_name: str = "address",
        debug: bool = False,
    ):
        self.parquet_handler = Parquet(debug=debug)
        if not os.path.exists(folder):
            raise ValueError(f"Invalid folder {folder}")
        self.folder = folder
        self.table_name = table_name
        if column_mapping is None:
            column_mapping = {
                "id": "id",
                "page": "page",
                "lastname": "lastname",
                "firstname": "firstname",
                "Beruf o. ä.": "occupation",
                "Straße": "street",
                "Ortsname": "location",
                "Ortskennung": "location_code",
                "abw./zus. Ortsangabe": "location_extra",
                "Firmenname": "company_name",
                "Familienstand": "marital_status",
                "Vorname Bezugsperson": "reference_person_firstname",
                "Beruf Bezugsperson": "reference_person_occupation",
                "Eigentümer": "owner",
                "Funktionsträger": "official",
                "abweichender Wohnort": "alternate_residence",
                "Bezirk": "district",
                "source": "source",
                "jahr": "year",
            }
        self.year_mapping = {"weimarTH1851.parquet": 1851, "weimarTH1853.parquet": 1853}
        self.column_mapping = column_mapping

    def inject_year(self):
        # Add year to each row based on the year_mapping
        for _table_name, rows in self.parquet_data.items():
            for row in rows:
                source = row.get("source")
                if source in self.year_mapping:
                    row["jahr"] = self.year_mapping[source]
                else:
                    print(f"Warning: No year mapping found for source {source}")

    def read(self):
        # Read all parquet files
        self.parquet_data = self.parquet_handler.read_parquet_files(self.folder)
        return self.parquet_data

    def convert(self, db: SQLDB):
        # Convert to SQLite with the correct table name and column mapping
        self.parquet_handler.convert_parquet_to_sqlite(
            self.parquet_data,
            db=db,
            table_name=self.table_name,
            column_mapping=self.column_mapping,
        )

    def to_db(self, db: SQLDB):
        self.read()
        self.inject_year()
        self.convert(db)



"""
Created on 2024-08-25

@author: wf
"""

import json

from ez_wikidata.wdsearch import WikidataSearch

from genwiki.gov_api import GOV_API
from genwiki.locator import Locator
from genwiki.nominatim import NominatimWrapper
from genwiki.wikidata import Wikidata
from tests.gbasetest import GenealogyBasetest


class TestLocations(GenealogyBasetest):
    """
    test location handling
    """

    def setUp(self, debug=False, profile=True):
        GenealogyBasetest.setUp(self, debug=debug, profile=profile)
        self.locator = Locator(debug=debug)

    def get_parts(self, items):
        parts = {}
        qlod = self.locator.multi_item_query(
            query_name="WikidataItemHasParts", items=items, lang="de"
        )
        for record in qlod:
            part = record["part"]
            part = Wikidata.unprefix(part)
            parts[part] = record
        return parts

    def test_location_hierarchy(self):
        """
        get a lookup of european locations
        """
        europe = "Q46"
        parts = self.get_parts([europe])
        sub_parts = self.get_parts(list(parts.keys()))
        part_to_item = {}
        for qid, record in sub_parts.items():
            record["item"] = Wikidata.unprefix(record["item"])
            record["part"] = Wikidata.unprefix(record["part"])
            item = record["item"]
            part = record["part"]

            print(f"{qid}:{record}")
            item_label = record["itemLabel"]
            part_label = record["partLabel"]

            # If the country is not already in the mapping, add it
            if part not in part_to_item:
                part_to_item[part] = record
            else:
                print(f"Duplicate assignment found for country {part_label}. ")

        # Export the mapping to a JSON file
        with open("/tmp/region_lookup.json", "w", encoding="utf-8") as f:
            json.dump(part_to_item, f, ensure_ascii=False, indent=4)

    def test_coords(self):
        """
        test getting coordinates for items
        """
        items = ["Q1729", "Q7070"]
        coords = self.locator.get_coordinates(items)
        expected = {
            "Q1729": (50.978055555, 11.028888888),  # Erfurt
            "Q7070": (50.974722222, 10.324444444),  # Eisenach
        }
        self.assertEqual(expected, coords)

    def test_path(self):
        """
        test getting a path
        """
        item = "Q3955"
        page_title = self.locator.lookup_path_for_item(item)
        self.assertEqual("DE/TH/Weimar", page_title)

    def test_sort_items(self):
        """
        test sorting items
        """
        items = {"a": "Q20", "b": "Q20", "c": "Q20", "d": "Q300", "e": "Q200"}
        expected = {"a": "Q20", "b": "Q20", "c": "Q20", "e": "Q200", "d": "Q300"}
        self.locator.sort_items(items)
        self.assertEqual(expected, items)

    def testLocator(self):
        """
        test the locator
        """
        geo_names_id = "3092080"
        qid = self.locator.lookup_wikidata_id_by_geoid(
            geoid_kind="geonames", geo_id=geo_names_id, lang="de"
        )
        self.assertEqual("Q255385", qid)

        debug = True
        for gov_id, expected in [
            ("RUMURGJO84LA", {"gov-Miastko@pl": "Q255385"}),
            (
                "adm_136611",
                {
                    "gov-Kassel@de": "Q2865",
                    "NUTS2003:DE731": "Q2865",
                    "NUTS1999:DE731": "Q2865",
                    "gov-Cassel@de": "Q11160",
                    "geonames:6547484": "Q32064001",
                },
            ),
            (
                "WEIMARJO50QX",
                {
                    "NUTS2003:DEG05": "Q3955",
                    "gov-Weimar@de": "Q3955",
                    "NUTS1999:DEG05": "Q3955",
                    "geonames:2812482": "Q32067990",
                },
            ),
            ("Vaihingen auf den Fildern", {}),
        ]:
            items = self.locator.locate(gov_id)
            if debug:
                print(f"{gov_id}:{items}")
            self.assertEqual(expected, items)

    def testWikidataSearch(self):
        """ """
        examples = [
            ("Q57993", "Thalheim", "Erzgebirgskreis"),
            ("Q47554", "Thorn", "Weichsel"),
            ("Q1123987", "Schloss Rundale", "Lettland"),
            ("Q193474", "Krain", "Herzogtum"),
        ]
        wds = WikidataSearch()
        limit = 10
        debug = self.debug
        debug = True
        found_location = None
        for i, (expected_qid, location, located_in) in enumerate(examples):
            wds.language = "de"
            sr = wds.searchOptions(location, limit=limit)
            if debug:
                print(f"{i:2}:{location}:{len(sr)}")
                print(json.dumps(sr, indent=2))
            for _j, record in enumerate(sr):
                qid, qlabel, desc = record
                if located_in in desc or located_in in qlabel:
                    found_location = record
        if found_location:
            qid, qlabel, desc = found_location
            if debug:
                print(f"found: {qid} {qlabel}-{desc}")
                self.assertEqual(expected_qid, qid)
        self.assertIsNotNone(found_location)

    def testGOV(self):
        """
        test gov api access
        """
        # Example usage
        gov_id = "WEIMARJO50QX"
        gov_api = GOV_API()
        obj = gov_api.get_raw_gov_object(gov_id)
        debug = self.debug
        debug = True
        if debug:
            print(json.dumps(obj, indent=2))

    def testNominatim(self):
        """
        test Nominatim functionality
        """
        nominatim = NominatimWrapper(user_agent="GenWiki2024LocationTest")
        examples = [
            ("Q57993", "Thalheim (Erzgebirge)"),
        ]
        debug = self.debug
        debug = True
        for expected_qid, location in examples:
            wikidata_id = nominatim.lookup_wikidata_id(location)
            self.assertIsNotNone(wikidata_id, f"Wikidata ID not found for {location}")
            self.assertEqual(
                expected_qid, wikidata_id, f"Unexpected Wikidata ID for {location}"
            )
            if debug:
                print(f"Wikidata ID for {location}: {wikidata_id}")



"""
Created on 26.08.2024

@author: wf
"""

from lodstorage.sparql import SPARQL
from lodstorage.sql import SQLDB
from ngwidgets.lod_grid import ListOfDictsGrid
from ngwidgets.widgets import Link
from nicegui import background_tasks, run, ui
from nicegui.events import ValueChangeEventArguments

from genwiki.multilang_querymanager import MultiLanguageQueryManager
from genwiki.params_view import ParamsView
from genwiki.wiki import Wiki
from genwiki.wikidata import Wikidata


class QueryView:
    """
    simplified version of the snapquery version snapquery_view
    """

    def __init__(
        self, solution, mlqm: MultiLanguageQueryManager, sql_db: SQLDB, wiki: Wiki
    ):
        self.solution = solution
        self.mlqm = mlqm
        self.wiki = wiki
        self.sql_db = sql_db
        self.load_task = None
        self.timeout = 5.0
        self.params_view = None

    def setup_ui(self,query_name:str="Gesamtanzahl"):
        """
        setup the user interface
        """
        self.query_name = query_name
        with ui.row() as self.query_row:
            self.query_select = self.solution.add_select(
                "Abfrage",
                self.mlqm.query_names,
                on_change=self.on_update_query,
            ).bind_value(self, "query_name")
            self.refresh_button = ui.button(
                icon="refresh",
                on_click=self.on_update_query,
            ).tooltip("Parameter-Ansicht aktualisieren")
            self.quer_button = ui.button(
                icon="play_circle",
                on_click=self.run_query,
            ).tooltip("ausführen")
        with ui.row() as self.params_row:
            pass
        self.grid_row = ui.row()
        pass

    async def on_update_query(self, _vcae: ValueChangeEventArguments):
        """
        react on a changed query
        """
        self.query = self.mlqm.query4Name(self.query_name)
        if self.params_view:
            self.params_view.delete()
        if self.query and self.query.params and self.query.params.has_params:
            self.query.set_default_params(self.query.params.params_dict)
            self.params_view = ParamsView(
                solution=self.solution, params=self.query.params
            )
            self.params_view.setup()
            self.params_view.open()
        self.params_row.update()

    def get_query_lod(self):
        """
        run the query
        """
        query = self.query
        if query.lang == "sql":
            qlod = self.sql_db.query(query.query)
        elif query.lang == "sparql":
            sparql = Wikidata.get_sparql()
            qlod = sparql.queryAsListOfDicts(
                query.query, param_dict=query.params.params_dict
            )
        elif query.lang == "ask":
            qlod = self.wiki.query_as_list_of_dicts(query.query)
        else:
            raise ValueError(f"query language {query.lang} not supported")
        return qlod

    async def load_query_results(self):
        """
        (re) load the query results
        """
        try:
            if self.query.params and self.query.params.has_params:
                self.query.query = self.query.params.apply_parameters()
                self.params_view.close()
            lod = await run.io_bound(self.get_query_lod)
            if not lod:
                with self.solution.container:
                    ui.notify("query execution failure")
                return
            self.grid_row.clear()
            with self.query_row:
                record_count = len(lod) if lod is not None else 0
                # markup = f'<span style="color: green;">{record_count} records in {stats.duration:.2f} secs</span>'
            # tablefmt = "html"
            # self.query.preFormatWithCallBacks(lod, tablefmt=tablefmt)
            # self.query.formatWithValueFormatters(lod, tablefmt=tablefmt)
            for record in lod:
                for key, value in record.items():
                    if isinstance(value, str):
                        if value.startswith("http"):
                            record[key] = Link.create(value, value)
            with self.grid_row:
                self.lod_grid = ListOfDictsGrid()
                self.lod_grid.load_lod(lod)
            self.lod_grid.sizeColumnsToFit()
            self.grid_row.update()
        except Exception as ex:
            self.solution.handle_exception(ex)

    async def run_query(self, _args):
        """
        run the current query
        """

        def cancel_running():
            if self.load_task:
                self.load_task.cancel()

        self.grid_row.clear()
        with self.grid_row:
            ui.spinner()
        self.grid_row.update()
        # cancel task still running
        cancel_running()
        # cancel task if it takes too long
        ui.timer(self.timeout, lambda: cancel_running(), once=True)
        # run task in background
        self.load_task = background_tasks.create(self.load_query_results())
